///////////////////////////////////////////////////////////////
// 3270 Screen buffer object: 3270 Commands and Orders
///////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////
// The following include is needed because we define a
// critical section object to force single threading when
// writing error messages to a message log.
///////////////////////////////////////////////////////////////

#include <afxmt.h>

///////////////////////////////////////////////////////////////
// 3270 cells can contain either a field marker or a character
///////////////////////////////////////////////////////////////

typedef enum {
        CICS_3270_CELL_CHARACTER,  // Cell type - character 
        CICS_3270_CELL_FIELD       // Cell type - field     
} CICS_3270_CellEntry_t;

///////////////////////////////////////////////////////////////
// The following is the definition of a 3270 cell entry
///////////////////////////////////////////////////////////////

typedef struct CICS_3270_Cell
{
        CICS_3270_CellEntry_t   cell_type;
        char                    character;
        char                    field_attribute;
        char                    foregnd;
        char                    backgnd;
        char                    ext_hilite;
        char                    field_validate;
        char                    fld_outline;
        char                    ps_char_set;
        char                    transparency;
		char					SOSI;
} CICS_3270_Cell_t;

///////////////////////////////////////////////////////////////
// The following is the definition of a 3270 screen
///////////////////////////////////////////////////////////////

typedef struct CICS_3270_Screen
{
   int  rows;           // number of rows on screen     
   int  cols;           // number of columns on screen  
   int  altrows;        // number of alternate rows     
   int  altcols;        // number of alternate columns  
   int  fields;         // number of fields defined     
   int  unprot_fields;  // number of unprotected fields 
   int  first_attr;     // first attribute byte         
   int  last_attr;      // last attribute byte          
   int  first_unprot;   // first unprotected attribute  
   int  last_unprot;    // last unprotected attribute   
} CICS_3270_Screen_t;

///////////////////////////////////////////////////////////////
// The following is the definition of a 3270 field
///////////////////////////////////////////////////////////////

typedef struct CICS_3270_Field
{
        int     row;                    // Row 0-24
        int     column;                 // Column 0-79                                                  
        char    attribute;              // Field attribute 
		char	extattr;				// Extended field attribute 
		char	color;					// Color attribute 
        int     length;                 // Length of data                                               
        char    *data;                  // Data                                                                 
} CICS_3270_Field_t;

typedef enum {
        CICS_3270_CODING_EBCDIC_T,              // EBCDIC Telnet
        CICS_3270_CODING_ASCII                  // ASCII
} CICS_3270_Coding_t;

typedef enum {
        CICS_3270_TRACE_STATUS_INQ,
        CICS_3270_TRACE_STATUS_ON,
        CICS_3270_TRACE_STATUS_APPEND,
        CICS_3270_TRACE_STATUS_OFF,
		CICS_3270_TRACE_DATE_ON,
		CICS_3270_TRACE_DATE_OFF
} CICS_3270_TRACE_REQ;

// 3270 screen field characteristics table 

typedef struct _FIELD_STRUCT
{
	int fldlen;						// bytes to next field 
	int bufofs;						// offset in screen buffer 
	int fldrow;						// row 
	int fldcol;						// column 
	char fldtype;					// field type 
} FIELD_STRUCT;

typedef struct _FIELD_STRUCT* pFIELD_STRUCT;

///////////////////////////////////////////////////////////////
//
// AID definitions
//
///////////////////////////////////////////////////////////////

#define CICS_3270_AID_ENTER             0x27    // ENTER key 
#define CICS_3270_AID_CLEAR             0x5f    // CLEAR key 

#define CICS_3270_AID_PA1               0x25    // PA1 key 
#define CICS_3270_AID_PA2               0x3e    // PA2 key 
#define CICS_3270_AID_PA3               0x2c    // PA3 key 

#define CICS_3270_AID_PF1               0x31    // PF1 key 
#define CICS_3270_AID_PF2               0x32    // PF2 key 
#define CICS_3270_AID_PF3               0x33    // PF3 key 
#define CICS_3270_AID_PF4               0x34    // PF4 key 
#define CICS_3270_AID_PF5               0x35    // PF5 key 
#define CICS_3270_AID_PF6               0x36    // PF6 key 
#define CICS_3270_AID_PF7               0x37    // PF7 key 
#define CICS_3270_AID_PF8               0x38    // PF8 key 
#define CICS_3270_AID_PF9               0x39    // PF9 key 
#define CICS_3270_AID_PF10              0x3a    // PF10 key 
#define CICS_3270_AID_PF11              0x23    // PF11 key 
#define CICS_3270_AID_PF12              0x40    // PF12 key 
#define CICS_3270_AID_PF13              0x41    // PF13 key 
#define CICS_3270_AID_PF14              0x42    // PF14 key 
#define CICS_3270_AID_PF15              0x43    // PF15 key 
#define CICS_3270_AID_PF16              0x44    // PF16 key 
#define CICS_3270_AID_PF17              0x45    // PF17 key 
#define CICS_3270_AID_PF18              0x46    // PF18 key 
#define CICS_3270_AID_PF19              0x47    // PF19 key 
#define CICS_3270_AID_PF20              0x48    // PF20 key 
#define CICS_3270_AID_PF21              0x49    // PF21 key 
#define CICS_3270_AID_PF22              0x5b    // PF22 key 
#define CICS_3270_AID_PF23              0x2e    // PF23 key 
#define CICS_3270_AID_PF24              0x3c    // PF24 key 

#define CICS_3270_AID_OPID              0x57    // OPERID or MSR 
#define CICS_3270_AID_MSRE              0x58    // Extended (standard) MSR 
#define CICS_3270_AID_TRIG              0x22    // Trigger field    
#define CICS_3270_AID_PEN               0x3d    // SELECTOR PEN or CURSOR SELECT key 
#define CICS_3270_AID_CLRP              0x7c    // Clear Partition 
#define CICS_3270_AID_STRF              0x68    // Structured field 
#define CICS_3270_AID_NULL              0x00    // NULL field

///////////////////////////////////////////////////////////////
// 3270 commands 
///////////////////////////////////////////////////////////////

#define CICS_3270_CMD_W                 0x31    // Write 
#define CICS_3270_CMD_EW                0x35    // Erase/Write 
#define CICS_3270_CMD_EWA               0x3D    // Erase/Write Alternate 
#define CICS_3270_CMD_RB                0x32    // Read Buffer 
#define CICS_3270_CMD_RM                0x36    // Read Modified 
#define CICS_3270_CMD_RMA               0x3E    // Read Modified All 
#define CICS_3270_CMD_EAU               0x3F    // Erase All Unprotected 
#define CICS_3270_CMD_WSF               0x88    // Write Structured Field (not supported) 

#define CICS_3270_CMD_HELP              0x30    // InfOD help command 

///////////////////////////////////////////////////////////////
// 3270 EBCDIC commands - local non-SNA 
///////////////////////////////////////////////////////////////

#define CICS_3270_CMD_NONSNA_EBCDIC_W   0x01    // Write 
#define CICS_3270_CMD_NONSNA_EBCDIC_EW  0x05    // Erase/Write 
#define CICS_3270_CMD_NONSNA_EBCDIC_EWA 0x0D    // Erase/Write Alternate 
#define CICS_3270_CMD_NONSNA_EBCDIC_RB  0x02    // Read Buffer 
#define CICS_3270_CMD_NONSNA_EBCDIC_RM  0x06    // Read Modified 
#define CICS_3270_CMD_NONSNA_EBCDIC_RMA 0x0E    // Read Modified All 
#define CICS_3270_CMD_NONSNA_EBCDIC_EAU 0x0F    // Erase All Unprotected 
#define CICS_3270_CMD_NONSNA_EBCDIC_WSF 0x11    // Write Structured Field 

///////////////////////////////////////////////////////////////
// 3270 EBCDIC commands - SNA 
///////////////////////////////////////////////////////////////

#define CICS_3270_CMD_EBCDIC_W          0xF1    // Write 
#define CICS_3270_CMD_EBCDIC_EW         0xF5    // Erase/Write 
#define CICS_3270_CMD_EBCDIC_EWA        0x7E    // Erase/Write Alternate 
#define CICS_3270_CMD_EBCDIC_RB         0xF2    // Read Buffer 
#define CICS_3270_CMD_EBCDIC_RM         0xF6    // Read Modified 
#define CICS_3270_CMD_EBCDIC_RMA        0x6E    // Read Modified All 
#define CICS_3270_CMD_EBCDIC_EAU        0x6F    // Erase All Unprotected 
#define CICS_3270_CMD_EBCDIC_WSF        0xF3    // Write Structured Field 

#define CICS_3270_CMD_HELP              0x30    // InfOD help command 


///////////////////////////////////////////////////////////////
// Write Control Character (WCC) bit definitions 
///////////////////////////////////////////////////////////////

#define CICS_3270_WCC_RESET             0x40    // Reset 
#define CICS_3270_WCC_PWIDTH            0x30    // Printer width 
#define CICS_3270_WCC_PRINT             0x08    // Start printer 
#define CICS_3270_WCC_ALARM             0x04    // Sound alarm 
#define CICS_3270_WCC_FREEKB            0x02    // Keyboard restore 
#define CICS_3270_WCC_FRSET             0x01    // Reset modified data tags 

#define CICS_3270_WCC_P132              0x00    // Use 132 cols 
#define CICS_3270_WCC_P40               0x10    // Use 40 cols 
#define CICS_3270_WCC_P64               0x20    // Use 64 cols 
#define CICS_3270_WCC_P80               0x30    // Use 80 cols 

///////////////////////////////////////////////////////////////
// Order codes 
///////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////
// Please note that the modify field command was changed from   
// 0x1E to 0x1A to avoid a collision with the Field Mark format 
// control order, which is assigned the same value.  However,   
// some ASCII versions of CICS still use the old value.  You    
// must determine which value your CICS system uses and change  
// the value of the CICS_3270_ORDER_MF definition to match.     
// MF orders are generated when field attributes are changed    
// without erasing the screen.                                  
///////////////////////////////////////////////////////////////

#define CICS_3270_ORDER_SF              0x1D    // Start Field 
#define CICS_3270_ORDER_SFE             0x10    // Start Field Extended 
#define CICS_3270_ORDER_SBA             0x11    // Set Buffer Address 
#define CICS_3270_ORDER_SA              0x1F    // Set Attribute 
#define CICS_3270_ORDER_MF              0x1E    // Modify Field - old value 
#define CICS_3270_ORDER_MF_OLD          0x1E    // Modify Field - new value 
#define CICS_3270_ORDER_MF_NEW          0x1A    // Modify Field - new value 
#define CICS_3270_ORDER_IC              0x13    // Insert Cursor 
#define CICS_3270_ORDER_PT              0x09    // Program Tab 
#define CICS_3270_ORDER_RA              0x14    // Repeat to Address 
#define CICS_3270_ORDER_EUA             0x12    // Erase Unprot to Address 
#define CICS_3270_ORDER_GE              0x08    // Graphic Escape 

///////////////////////////////////////////////////////////////
// Attribute type codes 
///////////////////////////////////////////////////////////////

#define CICS_3270_TY_ALL                0x00    // All character attributes 
#define CICS_3270_TY_FA                 0xC0    // 3270 Field Attribute 
#define CICS_3270_TY_FOUTLINE           0xC2    // Field outlining 
#define CICS_3270_TY_HILIGHT            0x41    // Extended hilighting 
#define CICS_3270_TY_FORE               0x42    // Foreground color 
#define CICS_3270_TY_BACK               0x45    // Background color 
#define CICS_3270_TY_SOSI               0xFE    // SOSI chars allowed 
#define CICS_3270_TY_PS                 0x43    // Character set 

#define CICS_3270_TY_TRANSP             0x46    // Background transparency 
#define CICS_3270_TY_FVALIDATE          0xC1    // Field validation 

///////////////////////////////////////////////////////////////
// Attribute values 
///////////////////////////////////////////////////////////////

#define CICS_3270_VA_DEFAULT            0x00    // Default settings 

#define CICS_3270_VA_NORMAL             0x30    // Normal hilighting 
#define CICS_3270_VA_BLINK              0x31    // Blink hilighting 
#define CICS_3270_VA_REVERSE            0x32    // Reverse video hilighting 
#define CICS_3270_VA_UNDERSCORE         0x34    // Underscored hilighting 

#define CICS_3270_VA_TOP                0x04    // Top outline 
#define CICS_3270_VA_BOTTOM             0x01    // Bottom outline 
#define CICS_3270_VA_LEFT               0x08    // Left outline 
#define CICS_3270_VA_RIGHT              0x02    // Right outline 

#define CICS_3270_VA_BLUE               0x31    // Blue 
#define CICS_3270_VA_RED                0x32    // Red 
#define CICS_3270_VA_PINK               0x33    // Pink 
#define CICS_3270_VA_GREEN              0x34    // Green 
#define CICS_3270_VA_TURQUOISE          0x35    // Turquoise 
#define CICS_3270_VA_YELLOW             0x36    // Yellow 
#define CICS_3270_VA_WHITE              0x37    // White 

#define CICS_3270_VA_NEUTRAL    CICS_3270_VA_WHITE      // Neutral colour (White) 

#define CICS_3270_VA_PS8                0x38    // PS=8 value 
#define CICS_3270_VA_PSF8               0xf8    // PS=8 value in EBCDIC 

#define CICS_3270_VA_SOSI               0x01    // SOSI input allowed 

///////////////////////////////////////////////////////////////
// Bit definitions for Field Attributes 
///////////////////////////////////////////////////////////////

#define CICS_3270_ATTR_PROT             0x20    // Field is protected 
#define CICS_3270_ATTR_NUMERIC          0x10    // Field is numeric 
#define CICS_3270_ATTR_DISPLAY          0x0C    // Display attributes 
#define CICS_3270_ATTR_NORM             0x00    // Display/not sel-pen-det 
#define CICS_3270_ATTR_SEL              0x04    // Display/sel-pen-det 
#define CICS_3270_ATTR_HIGH             0x08    // Intensified/sel-pen-det 
#define CICS_3270_ATTR_INVIS            0x0C    // Nondisplay, nondet 
#define CICS_3270_ATTR_MDT              0x01    // Modified Data Tag 
#define CICS_3270_ATTR_ASKIP            (CICS_3270_ATTR_PROT | CICS_3270_ATTR_NUMERIC)

#define CICS_3270_ATTR_DEFAULTS         0x00    // Default field attributes 

///////////////////////////////////////////////////////////////
// Direction of command
///////////////////////////////////////////////////////////////
 
typedef enum {
        CICS_3270_DIRECTION_INBOUND,            // To CICS                              
        CICS_3270_DIRECTION_OUTBOUND            // From CICS                    
} CICS_3270_Direction_t;

// Levels of trace detail
#define CICS_3270_TRACE_MIN		0
#define CICS_3270_TRACE_SOME	1
#define CICS_3270_TRACE_MAX		2

// Types of fields to search for when finding a field 
// based on field number on the screen
#define CICS_3270_FIELDS_ALL	0
#define CICS_3270_FIELDS_UNPROT	1
#define CICS_3270_FIELDS_PROT	2

// define a specific user 3270 class for our application
// this class encapsulates the 3270 buffer and all 3270 processing

class C3270 : public CObject
{
public:
	C3270(); // standard constructor

	// Public methods
	void	Dump(CDumpContext& dc) const;
    void	CICS_3270_Initialize(int);
    void	CICS_3270_GetField(int, int, CICS_3270_Field_t **);
    void	CICS_3270_FreeField(CICS_3270_Field_t *field);
    void	CICS_3270_Parse(char *, int);
    void	CICS_3270_DumpBuffer(char, char);
    void	CICS_3270_GetCell(int, int, CICS_3270_Cell_t *);
    short	CICS_3270_ModifyField(int, int, const char *);
    void	CICS_3270_GetInbound(char, char **, int *);
    void	CICS_3270_GetBuffer(char, char **, int *);
    void	CICS_3270_ClearBuffer();
    void	CICS_3270_DumpFields();
    void	CICS_3270_PrintField(int rw, int cl);
    void	CICS_3270_SetCursor(int row, int col);
    short	CICS_3270_ChgTrace(CICS_3270_TRACE_REQ traceReq, const char *fileName);
	void	CICS_3270_SetTraceLevel(int tracelevel);
    void	CICS_3270_WriteTraceLine(char *traceLine, int traceLevel);
    short	CICS_3270_Keyin(const char *cmdData);
    void	CICS_3270_GetScreenValues(CICS_3270_Screen_t* scr);
    void	CICS_3270_GetFieldByNum(int num, CICS_3270_Field_t *fld);
    void	CICS_3270_GetDataAtLoc(int bufferOffset, int maxlength, char* data);
    void	CICS_3270_GetUnprotByNum(int num, CICS_3270_Field_t *fld);
    void	CICS_3270_GetFieldLocByNum(short num, short *row, short *col, int fldtype);
    void	CICS_3270_WriteFields();
    int		CICS_3270_ScreenSum();
    int		CICS_3270_getNumberFields();
    int		CICS_3270_getNumberUnprotFields();
    void	CICS_3270_GetScreenParms(short* rows, short* cols);
    void	CICS_3270_GetScreenImage(char* buff, char fieldCharProt, char fieldCharUnprot);
	void	CICS_3270_GetFieldAttrs(char* buff);
	void	CICS_3270_Write_Error_Log(char *msg);
	void	CICS_3270_GetEpiErrMsg(int rc, short TermIndex, char *msg);
	void	CICS_3270_GetProtTable(char *buff);

	// Instance variables to pass data in and out 
	int		cursorPosition;
	int		numberFields;
	int		numberUnprotFields;
	int		numberRows;
	int		numberColumns;

    private:

	// Internal Subroutines
	void	NextCell(int *buffer);
	void	PreviousCell(int *buffer);
	int		checkExtAttr(int index);
	void	buildAttrPairs(int bufferPtr, char *tempAttr);
	void	clearDataBuffer(int index);
	int		findFieldByNum(int num, int allfld);
	void	SetField(int buffer, CICS_3270_Field_t *fld);
	void	GetByNum(int num, CICS_3270_Field_t* fld, const int allfld);
	void	dumpDatastream(char *data, int length);
	void	findAttrChar(int *buffer);
	void	printWCC(char data);
	void	setExtAttr(char attrType, char attrVal, int buffer,
                CICS_3270_Cell_t *OutBuffer);
	void	CICS_3270_ClearMDTs();
	void	CICS_Next_Field(int *bufofs, int *attr);
	void	GetFields(FIELD_STRUCT fields[]);
	void	TabNextField(int *buffer, int direction);
	void	NextLocation(int *buffer);
	void	handleASKIP(int *buffer, int *setMDT);
	void	badKeyMsg(const char *ch);
	void	deleteKey(int buffer, int *setMDT);
	void	eraseInput();
	short	writeKeyStroke(char data, int *buffer, int *setMDT);

	// Private variables

///////////////////////////////////////////////////////////////
//
// Buffer for holding the 3270 data stream (inbound, TO CICS
// Index into data for next items.
//
///////////////////////////////////////////////////////////////
	unsigned char   InDataBuffer[256*256];
	int             InDataBufferIndex;

	CICS_3270_Cell_t    OutDataBuffer[43*132];
	int		OutDataBufferIndex;
	int		firstAttr;
	int		lastAttr;
	int		firstUnprot;
	int		lastUnprot;
	int		alternateRows;
	int		alternateColumns;
	int		DEBUG;
	int		DEBUG_TIME_STAMP;
	int		DEBUG_LEVEL;
	FILE	*DEBUG_FILE;
};

/////////////////
// End Of File //
/////////////////
