/////////////////////////////////////////////////////////
//
// File: c3270.cpp
// Author: Jim MacNair
// Version: 1.0
// Based on an earlier version by Neil Kolban
//
/////////////////////////////////////////////////////////

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "c3270.h"
#include "epimsg.h"

#define MAX_LINE   255          // Maximum length of command 

static char Version[]=\
"@(#)c3270 object - CICS 3270 datastream V1.0 - Neil Kolban/Jim MacNair ("__DATE__")";

// Critical Section class object for single threading error log requests
static CCriticalSection	err_cs;

int     EbcdicCommandTable[256];
int     EbcdicTelnetCommandTable[256];
int     AsciiCommandTable[256];

//
//
// 3270 Data stream manipulation functions
//
//
unsigned char Bits2Ascii[64]={
0x20, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
0x48, 0x49, 0x5b, 0x2e, 0x3c, 0x28, 0x2b, 0x21,
0x26, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
0x51, 0x52, 0x5d, 0x24, 0x2a, 0x29, 0x3b, 0x5e,
0x2d, 0x2f, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
0x59, 0x5a, 0x7c, 0x2c, 0x25, 0x5f, 0x3e, 0x3f,
0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
0x38, 0x39, 0x3a, 0x23, 0x40, 0x27, 0x3d, 0x22};

unsigned char Ascii2Bits[256];

//////////////////////////////////////////////////////////////////////
// Constructor for the 3270 Screen object
//////////////////////////////////////////////////////////////////////
// Create a separate object to encapsulate the CICS 3270 Screen
C3270::C3270() : CObject()

{
	// Initialize all the instance variables
	InDataBufferIndex = 0;
	OutDataBufferIndex = 0;
	alternateRows = 24;
	alternateColumns = 80;
	cursorPosition=0;
	DEBUG = 0;
	DEBUG_LEVEL = CICS_3270_TRACE_MAX;
	DEBUG_FILE = NULL;
	DEBUG_TIME_STAMP = 0;
	firstAttr = 0;
	lastAttr = 0;
	firstUnprot = 0;
	lastUnprot = 0;
	numberFields=0;
	numberUnprotFields=0;
	numberRows=24;
	numberColumns=80;
}

////////////////////////////////////////////////////////////////
// This routine is a sample of how to dump out the values of an 
// object, so that the afxdump facility can be used elsewhere in
// an application.
////////////////////////////////////////////////////////////////

void C3270::Dump(CDumpContext& dc) const

{
	// Call the base class function first, to dump the address
	// of the object
	CObject::Dump(dc);

	// Next, dump out the internal variables
	dc << "numberRows = " << numberRows << "  numberColumns = " << numberColumns << "\n";
	dc << "numberFields  = " << numberFields << "  ";
	dc << "numberUnprotFields  = " << numberUnprotFields << "\n";
	dc << "numberRows  = " << numberRows << "  ";
	dc << "numberColumns  = " << numberColumns << "\n";
	dc << "first field = " << firstAttr << "  last field = " << lastAttr << "\n";
	dc << "first Unprot = " << firstUnprot << "  last Unprot = " << lastUnprot << "\n";
	dc << "Cursor Position = " << cursorPosition << "\n";
}

////////////////////////////////////////////////////////////////
// Routine to convert a character to upper case
////////////////////////////////////////////////////////////////

char chrupr(char ch)

{
   char tstr[2];

   // convert the character to a string 
   tstr[0] = ch;
   tstr[1] = '\0';

   // translate to upper case 
   strupr(tstr);

   // return the translated character 
   return tstr[0];
}

////////////////////////////////////////////////////////////////
// func(): CICS_3270_GetScreenValues()
//
// This routine will return the characteristics of the
// current screen image.
//
////////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetScreenValues(CICS_3270_Screen_t *scr)

{
   scr->rows          = numberRows;
   scr->cols          = numberColumns;
   scr->altrows       = alternateRows;
   scr->altcols       = alternateColumns;
   scr->fields        = numberFields;
   scr->unprot_fields = numberUnprotFields;
   scr->first_attr    = firstAttr;
   scr->last_attr     = lastAttr;
   scr->first_unprot  = firstUnprot;
   scr->last_unprot   = lastUnprot;
}

//////////////////////////////////////////////////////////
//
// Subroutine to turn a 4-bit number into a hex character
//
//////////////////////////////////////////////////////////

void data2hex(char *ch)

{
   if (((*ch) >= 0) && ((*ch) <= 9))
   {
      (*ch) += 48;  // turn into number 0-9 
   }
   else
   {
      (*ch) += 55;  // turn into letter A-F 
   } // endif 
}

void C3270::NextCell(int *buffer)

{
   (*buffer)++;
   if (*buffer >= numberRows*numberColumns)
   {
      *buffer = 0;
   }

   return;
}

void C3270::PreviousCell(int *buffer)

{
   (*buffer)--;
   if (*buffer < 0)
   {
      *buffer = numberRows*numberColumns-1;
   }
}

/////////////////////////////////////////////////////////////////
//
// Routine to check for extended attribute bytes for a location
// in the screen buffer.  If found, it returns a 1, otherwise
// it returns a 0.
//
/////////////////////////////////////////////////////////////////

int C3270::checkExtAttr(int index)

{
   if ((OutDataBuffer[index].foregnd != CICS_3270_ATTR_DEFAULTS) ||
       (OutDataBuffer[index].backgnd != CICS_3270_ATTR_DEFAULTS) ||
       (OutDataBuffer[index].ext_hilite != CICS_3270_ATTR_DEFAULTS) ||
       (OutDataBuffer[index].field_validate != CICS_3270_ATTR_DEFAULTS) ||
       (OutDataBuffer[index].fld_outline != CICS_3270_ATTR_DEFAULTS) ||
       (OutDataBuffer[index].ps_char_set != CICS_3270_ATTR_DEFAULTS) ||
       (OutDataBuffer[index].transparency != CICS_3270_ATTR_DEFAULTS) ||
       (OutDataBuffer[index].SOSI != CICS_3270_ATTR_DEFAULTS))
   {
      return 1;
   }
   else
   {
      return 0;
   }
}

void updatePair(char *loc,
                int  *attr,
                unsigned char type,
                unsigned char value)

{
   // set the type and value pair into the string 
   loc[0] = type;
   loc[1] = value;

   // update the pointer 
   (*attr) += 2;
}

/////////////////////////////////////////////////////////////////
//
// Routine to build a string of attribute type/value pairs
// for a given location on the screen.  The location is
// assumed to be an attribute byte.
//
/////////////////////////////////////////////////////////////////

void C3270::buildAttrPairs(int bufferPtr, char *tempAttr)

{
   int attr = 0;

   if (OutDataBuffer[bufferPtr].cell_type == CICS_3270_CELL_FIELD)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_FA,
                 Bits2Ascii[OutDataBuffer[bufferPtr].field_attribute]);
   } // endif 

   if (OutDataBuffer[bufferPtr].foregnd !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_FORE,
                 OutDataBuffer[bufferPtr].foregnd);
   } // endif 

   if (OutDataBuffer[bufferPtr].backgnd !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_BACK,
                 OutDataBuffer[bufferPtr].backgnd);
   } // endif 

   if (OutDataBuffer[bufferPtr].ext_hilite !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_HILIGHT,
                 OutDataBuffer[bufferPtr].ext_hilite);
   } // endif 

   if (OutDataBuffer[bufferPtr].field_validate !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_FVALIDATE,
                 OutDataBuffer[bufferPtr].field_validate);
   } // endif 

   if (OutDataBuffer[bufferPtr].fld_outline !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_FOUTLINE,
                 OutDataBuffer[bufferPtr].fld_outline);
   } // endif 

   if (OutDataBuffer[bufferPtr].ps_char_set !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_PS,
                 OutDataBuffer[bufferPtr].ps_char_set);
   } // endif 

   if (OutDataBuffer[bufferPtr].transparency !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_TRANSP,
                 OutDataBuffer[bufferPtr].transparency);
   } // endif 

   if (OutDataBuffer[bufferPtr].SOSI !=
       CICS_3270_ATTR_DEFAULTS)
   {
      updatePair(tempAttr + attr, &attr, CICS_3270_TY_SOSI,
                 OutDataBuffer[bufferPtr].SOSI);
   } // endif 

   tempAttr[attr] = '\0';
}

/////////////////////////////////////////////////////////////////
//
// Routine to clear the attribute bytes for a location
// in the screen buffer
//
/////////////////////////////////////////////////////////////////

void C3270::clearDataBuffer(int index)

{
   OutDataBuffer[index].foregnd        = 0;
   OutDataBuffer[index].backgnd        = 0;
   OutDataBuffer[index].ext_hilite     = 0;
   OutDataBuffer[index].field_validate = 0;
   OutDataBuffer[index].fld_outline    = 0;
   OutDataBuffer[index].ps_char_set    = 0;
   OutDataBuffer[index].transparency   = 0;
   OutDataBuffer[index].SOSI           = 0;
}

/////////////////////////////////////////////////////////////////
//
// func(): CICS_3270_ClearBuffer()
//
// Clear the 3270 data stream buffer. Reset all attribute bytes and
// set the data values to a null (binary zero).
//
/////////////////////////////////////////////////////////////////

void C3270::CICS_3270_ClearBuffer()

{
   int     index;

   OutDataBufferIndex = 0;
   cursorPosition = 0;
   numberFields = 0;
   numberUnprotFields = 0;
   firstAttr = -1;
   lastAttr = -1;
   firstUnprot = -1;
   lastUnprot = -1;

   //
   // For each of the cell positions, set to NULL
   //
   for (index=0; index<numberRows*numberColumns; index++)
   {
      OutDataBuffer[index].cell_type   = CICS_3270_CELL_CHARACTER;
      OutDataBuffer[index].character   = 0;
      clearDataBuffer(index);
   }

   return;
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_Initialize()
//
// Initialize the 3270ds functions
//
/////////////////////////////////////////////////////////////////

void C3270::CICS_3270_Initialize(
/* IN */ int model)

{
   int     index;
   char    *debugEnv;

   //
   // Set the alternate row and column numbers
   //

   switch (model)
   {
       case 1:
          alternateRows      = 12;
          alternateColumns   = 40;
          break;

       case 2:
          alternateRows      = 24;
          alternateColumns   = 80;
          break;

       case 3:
          alternateRows      = 32;
          alternateColumns   = 80;
          break;

       case 4:
          alternateRows      = 43;
          alternateColumns   = 80;
          break;

       case 5:
          alternateRows      = 27;
          alternateColumns   = 132;
          break;

       default:
          alternateRows      = 24;
          alternateColumns   = 80;
          break;
   } // endswitch 

   numberRows      = 24;
   numberColumns   = 80;

   // Clear the 3270 buffer 
   CICS_3270_ClearBuffer();

   //
   // Build the ASCII to Buffer address table
   //
   for (index=0; index<64; index++)
   {
      Ascii2Bits[Bits2Ascii[index]] = (unsigned char)index;
   }

   EbcdicCommandTable[CICS_3270_CMD_EBCDIC_W]               = CICS_3270_CMD_W;
   EbcdicTelnetCommandTable[CICS_3270_CMD_NONSNA_EBCDIC_W]  = CICS_3270_CMD_W;
   AsciiCommandTable[CICS_3270_CMD_W]                       = CICS_3270_CMD_W;
   EbcdicCommandTable[CICS_3270_CMD_EBCDIC_EW]              = CICS_3270_CMD_EW;
   EbcdicTelnetCommandTable[CICS_3270_CMD_NONSNA_EBCDIC_EW] = CICS_3270_CMD_EW;
   AsciiCommandTable[CICS_3270_CMD_EW]                      = CICS_3270_CMD_EW;

   EbcdicCommandTable[CICS_3270_CMD_EBCDIC_EWA]              = CICS_3270_CMD_EWA;
   EbcdicTelnetCommandTable[CICS_3270_CMD_NONSNA_EBCDIC_EWA] = CICS_3270_CMD_EWA;
   AsciiCommandTable[CICS_3270_CMD_EWA]                      = CICS_3270_CMD_EWA;

   EbcdicCommandTable[CICS_3270_CMD_EBCDIC_RB]              = CICS_3270_CMD_RB;
   EbcdicTelnetCommandTable[CICS_3270_CMD_NONSNA_EBCDIC_RB] = CICS_3270_CMD_RB;
   AsciiCommandTable[CICS_3270_CMD_RB]                      = CICS_3270_CMD_RB;

   debugEnv = (char *)getenv("DEBUG_3270");
   if (debugEnv != NULL)
   {
      if (strcmp(debugEnv, "stdout") == 0)
      {
         DEBUG_FILE = stdout;
      }
      else
      {
         DEBUG_FILE = fopen(debugEnv, "w");
      }

      DEBUG = 1;
   }

   return;
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_ChgTrace()
//
// Routine to turn debug on or off, or print the current status,
// and to select a filename if desired.
//
/////////////////////////////////////////////////////////////////

short C3270::CICS_3270_ChgTrace(CICS_3270_TRACE_REQ traceReq, const char *fileName)

{
	char	currtime[20];	// work variable for ASCII time
	char	currdate[20];	// work variable for ASCII date

	// Get the current time and date
	_strdate(currdate);
	_strtime(currtime);

   if ((traceReq == CICS_3270_TRACE_STATUS_ON) ||
       (traceReq == CICS_3270_TRACE_STATUS_APPEND))
   {
      if ((DEBUG) && (DEBUG_FILE != NULL))
      {
         fclose(DEBUG_FILE);

		 // Clear the handle = no longer valid
		 DEBUG_FILE = NULL;
      }

      if (strlen(fileName) == 0)
      {
         DEBUG_FILE = stdout;
      }
      else
      {
         if (traceReq == CICS_3270_TRACE_STATUS_APPEND)
         {
            DEBUG_FILE = fopen(fileName, "a");
         }
         else
         {
            DEBUG_FILE = fopen(fileName, "w");
         }
      }

      if (DEBUG_FILE != NULL)
      {
         fprintf(DEBUG_FILE,"Trace is enabled at %s on %s\n",currtime,currdate);
         DEBUG = 1;
      }
   }

   if (traceReq == CICS_3270_TRACE_STATUS_OFF)
   {
		if (DEBUG_FILE != NULL)
		{
			fprintf(DEBUG_FILE,"Trace is disabled at %s on %s\n",currtime,currdate);

			// Close the file
			fclose(DEBUG_FILE);
			DEBUG_FILE = NULL;
		}

		DEBUG=0;
   }

   if (traceReq == CICS_3270_TRACE_DATE_ON)
   {
	   DEBUG_TIME_STAMP = 1;
   }

   if (traceReq == CICS_3270_TRACE_DATE_OFF)
   {
	   DEBUG_TIME_STAMP = 0;
   }

   // return the current trace status
   return DEBUG;
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_SetTraceLevel()
//
// Routine to turn debug on or off, or print the current status,
// and to select a filename if desired.
//
/////////////////////////////////////////////////////////////////

void C3270::CICS_3270_SetTraceLevel(int tracelevel)

{
	switch (tracelevel)
	{
	case CICS_3270_TRACE_MIN:
		DEBUG_LEVEL = CICS_3270_TRACE_MIN;
		break;

	case CICS_3270_TRACE_SOME:
		DEBUG_LEVEL = CICS_3270_TRACE_SOME;
		break;

	default:
		DEBUG_LEVEL = CICS_3270_TRACE_MAX;
		break;
	}
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_WriteTraceLine()
//
// Routine to write a debug line, usually from another
// module
//
/////////////////////////////////////////////////////////////////

void C3270::CICS_3270_WriteTraceLine(char *traceLine, int traceLevel)

{
	int		year;
	char	currtime[20];	// work variable for ASCII time
	char	currdate[20];	// work variable for ASCII date
	SYSTEMTIME	curtime;	// Work variable for current time

	// write a trace line, if trace is enabled, and force it to the disk
	if ((DEBUG) && (traceLevel <= DEBUG_LEVEL))
	{
		if (DEBUG_TIME_STAMP)
		{
			// Get the current time and date
			GetLocalTime(&curtime);

			// Format the date and and time
			year = curtime.wYear%100;
			sprintf(currdate, "%2.2d/%2.2d/%2.2d", year, curtime.wMonth, curtime.wDay);
			sprintf(currtime, "%2.2d:%2.2d:%2.2d.%3.3d", curtime.wHour, curtime.wMinute, 
					curtime.wSecond, curtime.wMilliseconds);

			// Check if the first character is a newline character
			if (traceLine[0] == '\n')
			{
				// Write the trace record
				fprintf(DEBUG_FILE,"\n%s %s %s",currdate,currtime,traceLine + 1);
			}
			else
			{
				// Write the trace record
				fprintf(DEBUG_FILE,"%s %s %s",currdate,currtime,traceLine);
			}
		}
		else
		{
			fprintf(DEBUG_FILE,"%s",traceLine);
		}

		// force the data out to disk
		fflush(DEBUG_FILE);
	} // endif 
}

///////////////////////////////////////////////////////////////
// Constant to define the name of a default error log file
///////////////////////////////////////////////////////////////

#define ERR_FILE_NAME "err3270.log"

///////////////////////////////////////////////////////////////
// Routine to log errors - error logging is always
// enabled.  The default file name is "simerror.msg".
// This file name can be changed if desired.
// This routine is intentionally kept separate, to allow
// it to be replaced by some other error logging process.
// For example, if the main DLL is used only under IIS,
// the log file might be replaced by using the IIS log file.
// Another logical place for error logging is the NT event
// log.
// Since C3270 is generally multi-threaded, a critical 
// section is used to single thread error logging events.
///////////////////////////////////////////////////////////////

void C3270::CICS_3270_Write_Error_Log(char *msg)

{
	FILE*	ERR_FILE;
	char*	errorEnv;
	char*   err_file_name;
	char	currtime[20];	// work variable for ASCII time
	char	currdate[20];	// work variable for ASCII date

	// Get the current time and date
	_strdate(currdate);
	_strtime(currtime);

	// Lock to ensure the next section is single threaded
	err_cs.Lock();

	// Check for error log file name in environment variable
	errorEnv = (char *)getenv("ERROR_3270");
	if (errorEnv == NULL)
	{
		// Set to default
		err_file_name = ERR_FILE_NAME;
	}
	else
	{
		// Set to environment variable value
		err_file_name = errorEnv;
	}

	// Add the error to the error log
	ERR_FILE = fopen(err_file_name,"a");    // open error log for append 
	if (ERR_FILE != NULL)
	{
		// Check if the first character is a newline character
		if (msg[0] == '\n')
		{
			// Write the trace record
			fprintf(ERR_FILE,"\n%s %s %s",currdate,currtime,msg + 1);
		}
		else
		{
			// Write the trace record
			fprintf(ERR_FILE,"%s %s %s",currdate,currtime,msg);
		}

		fclose(ERR_FILE);					// Close the error log
	}

	// Release the lock
	err_cs.Unlock();
}

void C3270::CICS_3270_GetEpiErrMsg(int rc,
                            short TermIndex,
                            char *msg)

{
	GetEpiErrMsg(rc, TermIndex, msg);
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_SBA_BufferToAscii()
//
// Convert from 12 bit buffer address to 2 byte (short) ASCII character
// representation suitable for sending to and from CICS.
//
/////////////////////////////////////////////////////////////////

void CICS_3270_SBA_BufferToAscii(int buffer, unsigned short *ascii)

{
   char *ptr;
   ptr = (char *)ascii;
   ptr[0] = Bits2Ascii[((buffer & 0x0f00) >> 6) | ((buffer & 0x00c0) >> 6)];
   ptr[1] = Bits2Ascii[buffer & 0x003f];
   return;
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_SBA_AsciiToBuffer()
//
// Convert from 2 byte ASCII representation to a buffer address.
//
/////////////////////////////////////////////////////////////////

void CICS_3270_SBA_AsciiToBuffer(unsigned short ascii, int *buffer)
{
   char *ptr;
   ptr = (char *)&ascii;

   *buffer = Ascii2Bits[ptr[1]] | Ascii2Bits[ptr[0]] << 6;

   return;
}

/////////////////////////////////////////////////////////////////
//
// Subroutine to find the buffer offset of a field
// based on the field number.  The routine starts at
// the beginning of the buffer and searches until
// num fields have been found.  If the end of the buffer
// is reached before the field is found, then a -1
// is returned.
// If allfld is set to 0, then all fields are counted.
// If allfld is set to 1, then only unprotected fields are counted.
// If allfld is set to 2, then only protected fields are counted.
//
/////////////////////////////////////////////////////////////////

int C3270::findFieldByNum(int num, int allfld)

{
   int buffer = 0;   // buffer location      
   int i = 0;        // field number counter 

   while (buffer < numberRows*numberColumns)
   {
      if (OutDataBuffer[buffer].cell_type == CICS_3270_CELL_FIELD)
      {
         switch (allfld) 
		 {
		 case CICS_3270_FIELDS_ALL:

            i++;   // increment the field counter 
			break;

		 case CICS_3270_FIELDS_UNPROT:
            if ((OutDataBuffer[buffer].field_attribute & CICS_3270_ATTR_PROT) == 0)
	        {
				i++;	// increment the field number
			}
			break;

		 case CICS_3270_FIELDS_PROT:
            if ((OutDataBuffer[buffer].field_attribute & CICS_3270_ATTR_PROT) != 0)
	        {
				i++;	// increment the field number
			}
		 } // endswitch 

         // check if we have found the right field 
         if (i == num)
         {
            // return the buffer address 
            return buffer;
         } // endif 
      } // endif 

      buffer++;
   } // endwhile 

   return -1;
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_GetFieldLocByNum()
//
// Set the field to indicate no field was found
// The row and column are set to -1 to indicate that
// no field was found.
//
/////////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetFieldLocByNum(short num, short *row, short *col, int fldtype)

{
   int buffer;

   buffer = findFieldByNum(num, fldtype);	
   if (buffer == -1)
   {
      *row = -1;   // indicate invalid field number 
      *col = -1;   // indicate invalid field number 
   }
   else
   {
      *row = buffer/numberColumns;
      *col = buffer%numberColumns;
   }
}

/////////////////////////////////////////////////////////////////
//
// Set the field to indicate no field was found
// The row and column are set to -1 to indicate that
// no field was found.
//
/////////////////////////////////////////////////////////////////

void C3270::SetField(int buffer, CICS_3270_Field_t *fld)

{
   int  len=0;
   int  index;
   char tempChars[43*132];

   fld->row       = buffer/numberColumns;
   fld->column    = buffer%numberColumns;
   fld->attribute = OutDataBuffer[buffer].field_attribute;

   // point at the attribute byte 
   index = buffer;

   // now search for the next field 
   while(1)
   {
      NextCell(&index);
      if (OutDataBuffer[index].cell_type == CICS_3270_CELL_FIELD)
      {
         break;
      }

      tempChars[len] = OutDataBuffer[index].character;
      len++;
   }

   // terminate the data string 
   tempChars[len] = '\0';

   // return the field length 
   fld->length    = len;

   // acquire storage for the field data 
   fld->data = (char *) malloc(len + 1);

   // copy the field data to the acquired storage 
   memcpy(fld->data, tempChars, len);
   fld->data[len] = '\0';
}

/////////////////////////////////////////////////////////////////
//
// Set the field to indicate no field was found
// The row and column are set to -1 to indicate that
// no field was found.
//
/////////////////////////////////////////////////////////////////

void SetNoField(CICS_3270_Field_t *fld)

{
   fld->row       = -1;         // Invalid row      
   fld->column    = -1;         // Invalid column   
   fld->attribute = 0;          // No attribute     
   fld->length    = 0;          // Zero length      
   fld->data      = NULL;       // Null data string 
}

/////////////////////////////////////////////////////////////////
//
// Common subroutine to handle GetFieldByNum and
// GetUnprotByNum.
//
/////////////////////////////////////////////////////////////////

void C3270::GetByNum(int num, CICS_3270_Field_t* fld, const int allfld)

{
   int i=0;
   int buffer=0;

   // set the field to indicate no field found 
   SetNoField(fld);

   while (buffer < numberRows*numberColumns)
   {
      if (OutDataBuffer[buffer].cell_type == CICS_3270_CELL_FIELD)
      {
         if ((allfld) ||
             ((OutDataBuffer[buffer].field_attribute &
               CICS_3270_ATTR_PROT) == 0))
         {
            i++;   // increment the field counter 
         } // endif 

         // check if we have found the right field 
         if (i == num)
         {
            // set the field characteristics 
            SetField(buffer, fld);
            break;
         } // endif 
      } // endif 

      buffer++;
   } // endwhile 
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_GetFieldByNum()
//
// Find the nth field and return the
// characteristics of that field
//
/////////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetFieldByNum(int num, CICS_3270_Field_t* fld)


{
   GetByNum(num, fld, 1);
}

/////////////////////////////////////////////////////////////////
//
// CICS_3270_GetUnprotByNum()
//
// Find the nth unprotected field and return the
// characteristics of that field
//
/////////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetUnprotByNum(int num, CICS_3270_Field_t* fld)

{
   GetByNum(num, fld, 0);
}

////////////////////////////////////////////////////////////
//
// Routine to get maxLength characters of data, starting
// at bufferOffset.  If the request extends past the
// beginning of the screen, then the request will wrap
// around to the beginning of the screen.
//
////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetDataAtLoc(int bufferOffset, int maxlength, char* data)

{
	// output data index
	int		i=0;

	// limit the maximum to the largest buffer size supported
	if (maxlength > 27*132)
	{
		maxlength = 27*132 + 1;
	}

	// start getting the data
	while (maxlength > 0)
	{
		// check if we are past the end of the buffer
		if (bufferOffset > numberRows*numberColumns)
		{
			bufferOffset = 0;
		}

		// get the data value - if it is an attribute byte, substitute a blank
		// if the data value is less than a space, substitute a space
		if (OutDataBuffer[bufferOffset].cell_type == CICS_3270_CELL_FIELD)
		{
			data[i] = ' ';
		}
		else
		{
			if (OutDataBuffer[bufferOffset].character < ' ')
			{
				data[i] = ' ';
			}
			else
			{
				data[i] = OutDataBuffer[bufferOffset].character;
			}
		}

		// move on to the next screen buffer and output location
		i++;
		bufferOffset++;

		// decrement the length count
		maxlength--;
	}

	// Terminate the string 
	data[i] = '\0';
}

/////////////////////////////////////////////////////////////////
//
// dumpDatastream()
//
// Write out the raw 3270 datastream in hex and ascii
//
/////////////////////////////////////////////////////////////////

void C3270::dumpDatastream(char *data, int length)

{
   char         hexStr[41];
   char         asciiStr[41];
   char         ch;
   int          i;
   int          j;
   int          k;
   int          buffer=0;
   int          remaining;
   int          offset;
   int          count;

   // keep track of how much we have to do */
   remaining = length;

   while (remaining > 0)
   {
      if (remaining > 16)
      {
         count = 16;
      }
      else
      {
         count = remaining;
      } // endif 

      // initialize the output strings and buffer offset 
      offset = buffer;
      i = 0;    // initialize variable offset 
      j = 0;    // initialize variable offset 
      k = 0;    // initialize counter 

      while (k < count)
      {
         ch = (unsigned char) data[buffer] >> 4;
         data2hex(&ch);
         hexStr[i++] = ch;

         ch = (unsigned char) data[buffer] & 0x0F;
         data2hex(&ch);
         hexStr[i++] = ch;

         if ((data[buffer] > ' ') &&
             ((unsigned char) data[buffer] <= 127))
         {
            asciiStr[j++] = data[buffer];
         }
         else
         {
            asciiStr[j++] = '.';
         } // endif 

         // Go on to the next character 
         remaining--;
         buffer++;
         k++;   // increment counter 

         // Insert a blank for every fourth character 
         if (k%4 == 0)
         {
            hexStr[i++] = ' ';
         } // endif 
      } // endwhile 

      hexStr[i]   = '\0'; // Terminate the string 
      asciiStr[j] = '\0'; // Terminate the string 

      if (count > 0)
      {
         fprintf(DEBUG_FILE, "%.4d %-36s  %s\n",
                 offset,hexStr,asciiStr);
      }
   } // endwhile 
}

/////////////////////////////////////////////////////////////////
//
// printWCC()
//
// Print the bits of the Write Control Character
//
/////////////////////////////////////////////////////////////////

void C3270::printWCC(char data)

{
   char printline[100]="   WCC bits - ";

   if (data != 0)
   {
      if ((data & CICS_3270_WCC_RESET) > 0)
      {
         strcat(printline, "Reset  ");
      } // endif 

      if ((data & CICS_3270_WCC_PRINT) > 0)
      {
         strcat(printline, "Start Printer  ");
      } // endif

      if ((data & CICS_3270_WCC_ALARM) > 0)
      {
         strcat(printline, "Alarm  ");
      } // endif

      if ((data & CICS_3270_WCC_FREEKB) > 0)
      {
         strcat(printline, "Unlock Keyboard  ");
      } // endif

      if ((data & CICS_3270_WCC_FRSET) > 0)
      {
         strcat(printline, "Reset MDTs");
      } // endif

      fprintf(DEBUG_FILE, "%s\n",printline);
   } // endif
}

/////////////////////////////////////////////////////////////////
//
// exthi2string()
//
// Convert from extended highlight attribute value to string
//
/////////////////////////////////////////////////////////////////

void exthi2string(char data, char *str)

{
   switch (data)
   {
      case CICS_3270_VA_NORMAL:
         strcpy(str,"Normal\0");
         break;

      case CICS_3270_VA_BLINK:
         strcpy(str,"Blink\0");
         break;

      case CICS_3270_VA_REVERSE:
         strcpy(str,"Reverse\0");
         break;

      case CICS_3270_VA_UNDERSCORE:
         strcpy(str,"Underline\0");
         break;

      default:
         strcpy(str,"Unknown\0");
         break;
   }
}

/////////////////////////////////////////////////////////////////
//
// color2string()
//
// Convert from color attribute value to string
//
/////////////////////////////////////////////////////////////////

void color2string(char data, char *str)

{
   switch (data)
   {
      case CICS_3270_VA_BLUE:
         strcpy(str,"Blue\0");
         break;

      case CICS_3270_VA_RED:
         strcpy(str,"Red\0");
         break;

      case CICS_3270_VA_PINK:
         strcpy(str,"Pink\0");
         break;

      case CICS_3270_VA_GREEN:
         strcpy(str,"Green\0");
         break;

      case CICS_3270_VA_TURQUOISE:
         strcpy(str,"Turquoise\0");
         break;

      case CICS_3270_VA_YELLOW:
         strcpy(str,"Yellow\0");
         break;

      case CICS_3270_VA_WHITE:
         strcpy(str,"White\0");
         break;

      default:
         strcpy(str,"Unknown\0");
         break;
   }
}

/////////////////////////////////////////
//
// attr2str()
//
// Convert from attribute value to string
//
/////////////////////////////////////////

void attr2str(int i,char *protectfld)

{
	protectfld[0] = '\0';

	if ((i & CICS_3270_ATTR_PROT) > 0)
	{
		if ((i & CICS_3270_ATTR_NUMERIC) > 0)
		{
			strcat(protectfld,"ASKIP  ");
		}
		else
		{
			strcat(protectfld,"PROT   ");
		} // endif 
	}
	else
	{
		if ((i & CICS_3270_ATTR_NUMERIC) > 0)
		{
			strcat(protectfld,"NUM    ");
		}
		else
		{
			strcat(protectfld,"UNPROT ");
		}
	}

	if ((i & CICS_3270_ATTR_INVIS) == CICS_3270_ATTR_INVIS)
	{
		strcat(protectfld,"NONDISP ");
	}
	else
	{
		if ((i & CICS_3270_ATTR_HIGH) > 0)
		{
			strcat(protectfld,"HIGHINT ");
		}
	}

	if ((i & CICS_3270_ATTR_MDT) > 0)
	{
		strcat(protectfld,"MDT");
	}
}

///////////////////////////////////////////////////////////////////////////
//
// Routine to find the start of the field.  We do this by looking
// for a cell with type=field.  We start at this position and work
// backwards, wrapping from (0,0) to (79,23) if necessary.  The case where
// there are no fields will be detected and a fake field is returned
// (0,0)
//
///////////////////////////////////////////////////////////////////////////

void C3270::findAttrChar(int *buffer)

{
   int          startIndex;

   startIndex = (*buffer);
   while(1)
   {
   //
   // If we have found a field start, break the loop.
   //
      if (OutDataBuffer[*buffer].cell_type == CICS_3270_CELL_FIELD)
      {
         break;
      }

   //
   // back to the previous cell, wraping if necessary
   //
      PreviousCell(buffer);

   //
   // Check for the case where there are no fields ...
   //
      if (*buffer == startIndex)
      {
         (*buffer) = 0;
         break;
      }
   } // end while (1) 
}

//////////////////////////////////////////////////////////////////////////
//
// Routine to update the attributes of a field.
// for a cell with type=field.  We start at this position and work
// backwards, wrapping from (0,0) to (79,23) if necessary.  The case where
// there are no fields will be detected and a fake field is returned
// (0,0)
//
//////////////////////////////////////////////////////////////////////////

void C3270::setExtAttr(char attrType, char attrVal, int buffer,
                CICS_3270_Cell_t *OutBuffer)

{
   char            attribute;
   char            colorstr[12];
   char            attrstr[50];
   char            errline[MAX_LINE + 50];

   switch((unsigned char) attrType)
   {
      // Field Attribute 
      case CICS_3270_TY_FA:
         attribute = Ascii2Bits[attrVal];
         OutBuffer->field_attribute = attribute;

         attr2str(attribute,attrstr);
         if ((attribute & CICS_3270_ATTR_PROT) == 0)
         {
            // Count the unprotected fields 
            numberUnprotFields++;

            // keep track of first unprotected attribute 
            if ((firstUnprot == -1) || (buffer < firstUnprot))
            {
               firstUnprot = buffer;
            } // endif 

            // keep track of last unprotected attribute 
            if (buffer > lastUnprot)
            {
               lastUnprot = buffer;
            } // endif 
         } // endif 

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "    3270 Field: 0x%X (%s)\n",
               attribute,attrstr);
         }

         break;
   //
   // Foreground color:
   //
      case CICS_3270_TY_FORE:
         OutBuffer->foregnd = attrVal;

         if (DEBUG)
         {
            color2string(attrVal,colorstr);
            fprintf(DEBUG_FILE, "    Foreground: 0x%.2X %s\n",
                    attrVal,colorstr);
         }

         break;

   //
   // Background color:
   //
      case CICS_3270_TY_BACK:
         OutBuffer->backgnd = attrVal;

         if (DEBUG)
         {
            color2string(attrVal,colorstr);
            fprintf(DEBUG_FILE, "    Background: 0x%.2X %s\n",
                    attrVal,colorstr);
         }

         break;

   //
   // Extended Hilight - 0x41
   //
      case CICS_3270_TY_HILIGHT:
         OutBuffer->ext_hilite = attrVal;

         if (DEBUG)
         {
            exthi2string(attrVal,colorstr);
            fprintf(DEBUG_FILE, "    Extended Hilight: 0x%.2X %s\n",
                    attrVal,colorstr);
         }

         break;

   //
   // Field Outline
   //
      case CICS_3270_TY_FOUTLINE:
         OutBuffer->fld_outline = attrVal;

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "    Field Outline: 0x%.2X\n",
                    (attrVal));
         }

         break;

   //
   // Field Validation
   //
      case CICS_3270_TY_FVALIDATE:
         OutDataBuffer->field_validate = attrVal;

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "    Field Validation: 0x%.2X\n",
                    attrVal);
         }

         break;

   //
   // Program Symbols
   //
      case CICS_3270_TY_PS:
         OutDataBuffer->ps_char_set = attrVal;

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "    Prog Symbols: 0x%.2X\n",
                    (attrVal));
         }

         break;

   //
   // Field Transparency
   //
      case CICS_3270_TY_TRANSP:
         OutDataBuffer->transparency = attrVal;

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "    Transparency: 0x%.2X\n",
                    (attrVal));
         }

         break;

   //
   // SOSI allowed (DBCS support - Shift Out Shift In)
   //
      case CICS_3270_TY_SOSI:
         OutDataBuffer->SOSI = attrVal; 

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "    SOSI: 0x%.2X\n",
                    (attrVal));
         }

         break;

	//
    // Reset to defaults
    //
      case CICS_3270_TY_ALL:
         OutBuffer->foregnd        = 0;
         OutBuffer->backgnd        = 0;
         OutBuffer->ext_hilite     = 0;
         OutBuffer->field_validate = 0;
         OutBuffer->fld_outline    = 0;
         OutBuffer->ps_char_set    = 0;
         OutBuffer->transparency   = 0;
         OutBuffer->SOSI           = 0;

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, " Reset to Deflts: 0x%.2X\n",
                    (attrVal));
         }

         break;

   //
   // Unknown attribute value
   //

      default:
         // log the error 
         sprintf(errline, "Unexpected attribute: 0x%.2X\n",
                 attrType);
         CICS_3270_Write_Error_Log(errline);

         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "Unexpected attribute: 0x%.2X\n",
                    attrType);
         }

   }       // End of switch        
}

//////////////////////////////////////////////////
//
// func(): CICS_3270_ClearMDTs()
//
// This subroutine clears the Modified Data Tags.
//
//////////////////////////////////////////////////

void C3270::CICS_3270_ClearMDTs()

{
   int  i;
   int  endbuffer;

   endbuffer = numberRows*numberColumns;

   for (i=0; i < endbuffer; i++)
   {
      if (OutDataBuffer[i].cell_type == CICS_3270_CELL_FIELD)
      {
        //
        // Clear the MDT bit
        //
         OutDataBuffer[i].field_attribute &=
                (~CICS_3270_ATTR_MDT);
      }
   } // endfor 
}

//////////////////////////////////////////////////////
//
// func(): CICS_3270_Parse()
//
// Parse outbound 3270 data stream and hold in buffer.
//
//////////////////////////////////////////////////////

void C3270::CICS_3270_Parse(
/* IN */ char *data,                                    // 3270 data stream                                     
/* IN */ int length)                                    // length of data   
{
   //
   //         0          1           2 ...
   // +---------------+-----+---------------------+
   // | Write Command | WCC | Orders and Data ... |
   // +---------------+-----+---------------------+
   //
   //

   int             index;        // offset within 3270 datastream 
   unsigned short  ascii;
   int             buffer;
   int             pairs;
   int             i;
   int             endAddr;      // Used for EUA and RA processing 
   int             protFld;      // Used for EUA processing 
   int             lastByte=0;   // Used for PT processing 
   char            ch;
   char            attribute;
   char            attrstr[50];
   char            errline[MAX_LINE + 50];
   int				SBAcount = 0;	// number of Set Buffer Address orders 
   int				SFcount = 0;	// number of Start Field orders 
   int				SFEcount = 0;	// number of Start Field Extended orders 
   int				ICcount = 0;	// number of Insert Cursor orders 
   int				MFcount = 0;	// number of Modify Field orders 
   int				SAcount = 0;	// number of Set Attribute orders 
   int				RAcount = 0;	// number of Repeat to Address orders 
   int				EUAcount = 0;	// number of Erase Unprotected to Address orders 
   int				PTcount = 0;	// number of Program Tab orders 
   int				GEcount = 0;	// number of Graphic Escape orders 

   if (DEBUG)
   {
      sprintf(errline, "\nOutbound Data Stream Received - %d bytes\n", length);
	  CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
	  // Write out the raw data stream to the trace file
	  if (DEBUG_LEVEL == CICS_3270_TRACE_MAX)
	  {
			dumpDatastream(data, length);
	  }

      CICS_3270_WriteTraceLine("\nData Stream Interpreted as Follows\n", CICS_3270_TRACE_SOME);
   } // endif 

   switch(data[0])
   {
      // Write command 
      case CICS_3270_CMD_W:

         if (DEBUG && (DEBUG_LEVEL >= CICS_3270_TRACE_SOME))
         {
            fprintf(DEBUG_FILE, "CMD: Write");
         }

         // Clear the modified data tags if the            
         // FRSET bit is on in the Write Control Character 

         if ((data[1] & CICS_3270_WCC_FRSET) > 0)
         {
            CICS_3270_ClearMDTs();
         } // endif 

         break;

      // Erase Write command 
      case CICS_3270_CMD_EW:
         if (DEBUG && (DEBUG_LEVEL >= CICS_3270_TRACE_SOME))
         {
            fprintf(DEBUG_FILE, "CMD: Erase Write");
         }

         // set the number of rows and columns to default 
         numberRows = 24;
         numberColumns = 80;

         // clear the buffer, since it is an erase write 
         CICS_3270_ClearBuffer();

         break;

      // Erase Write Alternative command 
      case CICS_3270_CMD_EWA:
         if (DEBUG && (DEBUG_LEVEL >= CICS_3270_TRACE_SOME))
         {
            fprintf(DEBUG_FILE, "CMD: Erase Write Alternate");
         }

         // set the number of rows and columns to alternate 
         numberRows = alternateRows;
         numberColumns = alternateColumns;

         // clear the buffer, since it is an erase write 
         CICS_3270_ClearBuffer();

         break;

      default:
         // log the error 
         sprintf(errline, "*** Unrecognized Command: 0x%X\n", data[0]);
         CICS_3270_Write_Error_Log(errline);
         if (DEBUG)
         {
            fprintf(DEBUG_FILE, "*** Unexpected Command: 0x%X", data[0]);
         } // endif 
         return;
   }
   //
   // Parse the WCC
   //

	if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
	{
		fprintf(DEBUG_FILE, "    WCC: 0x%X\n", data[1]);
		printWCC(data[1]);       // print a formatted version 
	}

   for (index=2; index<length;)
   {
   //
   //    0       1,2      3
   // +-----+----------+
   // | SBA |  2 bytes |
   // +-----+----------+
   //
      switch (data[index])
      {
         case CICS_3270_ORDER_SBA:
			 // increment order counter 
			 SBAcount++;

            // last character was order or command 
            lastByte=0;

            ascii = *(short *)(data+index+1);
            index += 3;
            CICS_3270_SBA_AsciiToBuffer(ascii, &OutDataBufferIndex);

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
			{
               fprintf(DEBUG_FILE, "\nO> SBA: Ascii: 0x%X %d (Row:%d, Column:%d)\n",
                       ascii, OutDataBufferIndex, 
					   OutDataBufferIndex/numberColumns, 
					   OutDataBufferIndex%numberColumns);
               fprintf(DEBUG_FILE, "O> IC index = %d length = %d\n", index,length);
			}

            break;

         case CICS_3270_ORDER_IC:
			 // increment order counter 
			 ICcount++;

			 // last character was order or command 
            lastByte=0;

			// Calculate the cursor position
            cursorPosition = OutDataBufferIndex;

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
            {
               fprintf(DEBUG_FILE, "\nO> Insert Cursor at screen location %d\n",OutDataBufferIndex);
               fprintf(DEBUG_FILE, "O> IC index = %d\n", index);
            }

            index++;

            break;

   //
   //   +0       +1         +2     +3
   // +-----+------------+------+-------+-----+
   // | SFE | # of pairs | Type | Value | ... |
   // +-----+------------+------+-------+-----+
   //
   //   +0       +1         +2     +3
   // +-----+------------+------+-------+-----+
   // | MF  | # of pairs | Type | Value | ... |
   // +-----+------------+------+-------+-----+
   //

// Please note that the modify field command was changed from   
// 0x1E to 0x1A to avoid a collision with the Field Mark format 
// control order, which is assigned the same value.  However,   
// some ASCII versions of CICS still use the old value.  You    
// must determine which value your CICS system uses and change  
// the value of the CICS_3270_ORDER_MF definition to match.     
// This change is made in the C3270.H file.                     
// MF orders are generated when field attributes are changed    
// without erasing the screen.                                  

         case CICS_3270_ORDER_SFE:
         case CICS_3270_ORDER_MF_NEW:
         case CICS_3270_ORDER_MF_OLD:
            // handle start field extended and modify field 

            // last character was order or command 
            lastByte=0;

            if (data[index] == CICS_3270_ORDER_SFE)
            {
				 // increment order counter 
				 SFEcount++;

               OutDataBuffer[OutDataBufferIndex].cell_type =
                  CICS_3270_CELL_FIELD;
               OutDataBuffer[OutDataBufferIndex].field_attribute = 0;
                  clearDataBuffer(OutDataBufferIndex);

               numberFields++;  // Keep count of fields 

               // keep track of the first attribute in buffer 
               if ((firstAttr == -1) || (OutDataBufferIndex < firstAttr))
               {
                  firstAttr = OutDataBufferIndex;
               } // endif 

               // keep track of the last attribute in buffer 
               if (OutDataBufferIndex > lastAttr)
               {
                  lastAttr = OutDataBufferIndex;
               } // endif 

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
               {
                  fprintf(DEBUG_FILE, "\nO> SFE\n");
               }
            }
            else
            {
				 // increment order counter 
				 MFcount++;

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
               {
                  fprintf(DEBUG_FILE, "\nO> MF\n");
               }

               if (OutDataBuffer[OutDataBufferIndex].cell_type
                   != CICS_3270_CELL_FIELD)
               {
                  if (DEBUG)
                  {
                     fprintf(DEBUG_FILE, "\nO> MF rejected - no attribute\n");
                  }

                  index += 2 + (2 * data[index + 1]);
                  break;
               } // endif 
            } // endif 

            // Get count of attribute type and value pairs 
            pairs=data[index+1];

            // Point to first pair 
            index+=2;

            for (;pairs>0; pairs--)
            {
               setExtAttr(data[index], data[index + 1],
                          OutDataBufferIndex,
                          &OutDataBuffer[OutDataBufferIndex]);
               index+=2;  // Go on to next pair 
            } // End of pairs loop 

            NextCell(&OutDataBufferIndex);
            break;


   //
   //   +0       +1
   // +-----+-----------+------+
   // | SF  | Attribute | ...  |
   // +-----+-----------+------+
   //
         case CICS_3270_ORDER_SF:
			 // increment order counter 
			 SFcount++;

            // last character was order or command 
            lastByte=0;

            OutDataBuffer[OutDataBufferIndex].cell_type = CICS_3270_CELL_FIELD;
            OutDataBuffer[OutDataBufferIndex].field_attribute = 0;
            clearDataBuffer(OutDataBufferIndex);

            numberFields++;  // Keep count of fields 

            // keep track of the first attribute in buffer 
            if ((firstAttr == -1) || (OutDataBufferIndex < firstAttr))
            {
               firstAttr = OutDataBufferIndex;
            } // endif 

            // keep track of the last attribute in buffer 
            if (OutDataBufferIndex > lastAttr)
            {
               lastAttr = OutDataBufferIndex;
            } // endif 

            attribute = Ascii2Bits[data[index+1]];
            OutDataBuffer[OutDataBufferIndex].field_attribute = attribute;

            if ((attribute && CICS_3270_ATTR_PROT) == 0)
            {
               // Count the unprotected fields 
               numberUnprotFields++;

               // keep track of first unprotected attribute 
               if ((firstUnprot == -1) || (OutDataBufferIndex < firstUnprot))
               {
                  firstUnprot = OutDataBufferIndex;
               } // endif 

               // keep track of last unprotected attribute 
               if (OutDataBufferIndex > lastUnprot)
               {
                  lastUnprot = OutDataBufferIndex;
               } // endif 
            } // endif 

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
            {
               fprintf(DEBUG_FILE, "\nO> SF\n");
               i = attribute;
               attr2str(i,attrstr);
               fprintf(DEBUG_FILE, "    3270 Field: 0x%.2X (%s)\n",
                       attribute,attrstr);
            }   // if (DEBUG) 

            index+=2;
            NextCell(&OutDataBufferIndex);

            break;

         case CICS_3270_ORDER_PT:
			 // increment order counter 
			 PTcount++;

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
            {
               fprintf(DEBUG_FILE, "\nO> PT\n");
            }

            // find the next unprotected field             
            // if we reach the end of the buffer, then the 
            // buffer address will wrap to 0               

            while (OutDataBufferIndex < numberRows*numberColumns)
            {
               if ((OutDataBuffer[OutDataBufferIndex].cell_type
                    == CICS_3270_CELL_FIELD) &&
                   ((OutDataBuffer[OutDataBufferIndex].field_attribute
                    & CICS_3270_ATTR_PROT) == 0))
               {
                  break;
               }

			   NextCell(&OutDataBufferIndex);
            } // endwhile 

            // move past the attribute byte 
            NextCell(&OutDataBufferIndex);

            // check if we are following an order or command 
            if (lastByte == 1)
            {
               // following order or command - erase from 
               // here to the next attribute              
               while (1)
               {
                  if (OutDataBuffer[OutDataBufferIndex].cell_type
                      == CICS_3270_CELL_FIELD)
                  {
                     break;
                  }
                  else
                  {
                     // clear the data value and field attributes 
                     OutDataBuffer[OutDataBufferIndex].character = 0;
                     clearDataBuffer(OutDataBufferIndex);

                     // move on to the next location 
                     OutDataBufferIndex++;

                     // stop if end of buffer 
                     if (OutDataBufferIndex == numberRows*numberColumns)
                     {
                        OutDataBufferIndex = 0;
                        break;
                     } // endif 
                  } // endif 
               } // endwhile 
            }
            else
            {
               // last character was order or command 
               lastByte=0;
            } // endif 

			// Skip the program tab 
			index++;

            break;

         case CICS_3270_ORDER_RA:
         case CICS_3270_ORDER_EUA:
            // calculate the ending address 
            ascii = *(short *)(data+index+1);
            CICS_3270_SBA_AsciiToBuffer(ascii, &endAddr);

            // get the data fill character, write trace 
            // information and see if we are in an      
            // unprotected field (for EUA)              
            if (data[index] == CICS_3270_ORDER_RA)
            {
				 // increment order counter 
				 RAcount++;

               ch = data[index + 3];

		        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
				{
                  fprintf(DEBUG_FILE, "\nO> RA - char = %c",ch);
                  fprintf(DEBUG_FILE, "  from %d to %d\n",
                          OutDataBufferIndex, endAddr);
				}
            }
            else
            {
				 // increment order counter 
				 EUAcount++;

               ch = 0;   // fill character is null 
               if (numberFields == 0)
               {
                  // no fields - whole screen is 
                  // unprotected - set switch    
                  protFld = 0;
               }
               else
               {
                  findAttrChar(&buffer);
                  if ((OutDataBuffer[buffer].field_attribute &
                       CICS_3270_ATTR_PROT) == 0)
                  {
                     protFld = 0;
                  }
                  else
                  {
                     protFld = 1;
                  } // endif 
               } // endif 

		        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
				{
                  fprintf(DEBUG_FILE, "\nO> EUA");
                  fprintf(DEBUG_FILE, "  from %d to %d\n",
                          OutDataBufferIndex, endAddr);
				}
            } // endif 

            if (endAddr < numberRows*numberColumns)
            {
               while (OutDataBufferIndex != endAddr)
               {
                  if (OutDataBuffer[OutDataBufferIndex].cell_type
                      == CICS_3270_CELL_CHARACTER)
                  {
                     if (data[index] == CICS_3270_ORDER_RA)
                     {
                        OutDataBuffer[OutDataBufferIndex].character = ch;
                     }
                     else
                     {
                        if (protFld == 0)
                        {
                           OutDataBuffer[OutDataBufferIndex].character = ch;
                           clearDataBuffer(OutDataBufferIndex);
                        } // endif 
                     } // endif 
                  }
                  else
                  {
                     if (data[index] == CICS_3270_ORDER_RA)
                     {
                        OutDataBuffer[OutDataBufferIndex].character = ch;
                        OutDataBuffer[OutDataBufferIndex].cell_type =
                          CICS_3270_CELL_CHARACTER;
                        OutDataBuffer[OutDataBufferIndex].field_attribute = 0;
                        clearDataBuffer(OutDataBufferIndex);
                     }
                     else
                     {
                        if ((OutDataBuffer[OutDataBufferIndex].field_attribute &
                            CICS_3270_ATTR_PROT) == 0)
                        {
                           protFld = 0;
                        }
                        else
                        {
                           protFld = 1;
                        } // endif 
                     }
                  } // endif 

                  NextCell(&OutDataBufferIndex);
               } // endwhile 
            }
            else
            {
               if (DEBUG)
               {
                  fprintf(DEBUG_FILE, "\nO> Bad stop address %d\n",
                          endAddr);
               }
            } // endif 

			// Check if Repeat to Address 
            if (data[index] == CICS_3270_ORDER_RA)
            {
               index++;
            }

            // move past the command in the 3270 datastream 
            index += 3;
            break;

         case CICS_3270_ORDER_SA:
			 // increment order counter 
			 SAcount++;

            // last character was order or command 
            lastByte=0;

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
            {
               fprintf(DEBUG_FILE, "\nO> SA\n");
            }

            setExtAttr(data[index + 1], data[index + 2],
                       OutDataBufferIndex,
                       &OutDataBuffer[OutDataBufferIndex]);

            index += 3;  // Move past SA sequence 

            break;

         case CICS_3270_ORDER_GE:
			 // increment order counter 
			 GEcount++;

            // last character was order or command 
            lastByte=0;

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
            {
               fprintf(DEBUG_FILE,
                       "\nOrder received but not handled: 0x%.2X\n", data[index]);
            }

            // skip the GE order and character 
            index += 2;

            break;

   //
   // Regular character data, set the cell type to be character
   // and save the character.
   //
         default:
            // last character was data byte 
            lastByte=1;

            OutDataBuffer[OutDataBufferIndex].cell_type
                    = CICS_3270_CELL_CHARACTER;
            OutDataBuffer[OutDataBufferIndex].character = data[index];
            NextCell(&OutDataBufferIndex);

	        if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
            {
               fprintf(DEBUG_FILE, "%c", data[index]);
            }

            index++;

            // check if we are at the end of the buffer 
            if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX) && (index >= length))
            {
               fprintf(DEBUG_FILE,"\n");
            } // endif 

            break;

      }       // End switch 
   }       // End of data for loop 

   // At this point we must recalculate the number of fields, 
   // since some of the field attributes might have been at   
   // the same location.                                      
   numberFields = 0;
   numberUnprotFields = 0;

   for (i = 0; i < numberRows*numberColumns; i++)
   {
	   if (OutDataBuffer[i].cell_type == CICS_3270_CELL_FIELD)
	   {
		   numberFields++;
		   // check if the field is unprotected 
		   if ((OutDataBuffer[i].field_attribute & CICS_3270_ATTR_PROT) == 0)
		   {
			   numberUnprotFields++;
		   }
	   }
   }

   if ((DEBUG) && (DEBUG_LEVEL >= CICS_3270_TRACE_SOME))
   {
	   // write out the number of fields 
	   fprintf(DEBUG_FILE, "\nTotal number of fields %d  unprotected %d\n",
				numberFields, numberUnprotFields);

	   // write the heading 
	   fprintf(DEBUG_FILE, "Counts of 3270 orders found\n");

		// write out the 3270 order counters 
		if (SBAcount > 0)
		{
			fprintf(DEBUG_FILE, "SBA orders found - %d\n", SBAcount);
		}

		// write out the 3270 order counters 
		if (SFcount > 0)
		{
			fprintf(DEBUG_FILE, "SF  orders found - %d\n", SFcount);
		}

		// write out the 3270 order counters 
		if (SFEcount > 0)
		{
			fprintf(DEBUG_FILE, "SFE orders found - %d\n", SFEcount);
		}

		// write out the 3270 order counters 
		if (ICcount > 0)
		{
			fprintf(DEBUG_FILE, "IC  orders found - %d\n", ICcount);
		}

		// write out the 3270 order counters 
		if (MFcount > 0)
		{
			fprintf(DEBUG_FILE, "MF  orders found - %d\n", MFcount);
		}

		// write out the 3270 order counters 
		if (SAcount > 0)
		{
			fprintf(DEBUG_FILE, "SA  orders found - %d\n", SAcount);
		}

		// write out the 3270 order counters 
		if (RAcount > 0)
		{
			fprintf(DEBUG_FILE, "RA  orders found - %d\n", RAcount);
		}

		// write out the 3270 order counters 
		if (EUAcount > 0)
		{
			fprintf(DEBUG_FILE, "EUA orders found - %d\n", EUAcount);
		}

		// write out the 3270 order counters 
		if (PTcount > 0)
		{
			fprintf(DEBUG_FILE, "PT  orders found - %d\n", PTcount);
		}

		// write out the 3270 order counters 
		if (GEcount > 0)
		{
			fprintf(DEBUG_FILE, "GE orders found - %d\n", GEcount);
		}

		// If maximum level of tracing, write a screen image
		// to the trace file.
		if (DEBUG_LEVEL == CICS_3270_TRACE_MAX)
		{
			CICS_3270_DumpBuffer('@','~');
		}
   }

}

//////////////////////////////////////////////////////
//
// func(): CICS_Next_Field(int *bufofs, int *attr)
//
// Dump the location and characteristics of the next
// attribute byte.  The search starts one position
// after the bufofs field, and returns the next field
// position in bufofs.  If the end of the buffer is
// reached without finding an attribute byte, then a
// -1 is returned in the bufofs field.
//
//////////////////////////////////////////////////////

 void C3270::CICS_Next_Field(int *bufofs, int *attr)

{
   while (1)
   {
      // check if first time through and first position   
      // is an attribute.  If so, then return immediately 

      if ((*bufofs = 0) &&
         (OutDataBuffer[0].cell_type != CICS_3270_CELL_CHARACTER))
      {
         break;
      }

      // move to the next location in the buffer 
      NextCell(bufofs);

      // check if we are past the end of the buffer 
      if (*bufofs = 0)
      {
         *bufofs = -1;
         break;
      } // endif 

      // check if attribute - if it is, then return 
      if (OutDataBuffer[(*bufofs)].cell_type != CICS_3270_CELL_CHARACTER)
      {
         (*attr) = OutDataBuffer[(*bufofs)].cell_type;
         break;
      }

   } // endwhile 

}

////////////////////////////////////////////////////////////////////////
//
// func(): CICS_3270_DumpBuffer()
//
// Dump the contents of the virtual 3270 screen to the trace file in a
// formatted form.
//
////////////////////////////////////////////////////////////////////////

void C3270::CICS_3270_DumpBuffer(
/* IN */ char protChar,
/* IN */ char unprotChar)

{
   int     x,y;				// work variables 
   int     row;				// cursor position - row 
   int     column;			// cursor position - column 
   int     bufferIndex;		// location within the screen buffer 
   int     MDTcount = 0;	// number of fields with modified data tag set 
   char		scrline[134];	// screen line work variable

   if (DEBUG)
   {
		// Tell what we are doing
		CICS_3270_WriteTraceLine("3270 Screen image\n",CICS_3270_TRACE_MIN);

		// start with the first position 
		bufferIndex = 0;
		for (y=0; y<numberRows; y++)
		{
			for (x=0; x<numberColumns; x++)
			{
				if (OutDataBuffer[bufferIndex].cell_type ==
					CICS_3270_CELL_CHARACTER)
				{
					// check for characters less than a blank and
					// substitute a blank character
					// write the character to the screen image file 
					// note that field mark characters will changed to blanks
					scrline[x] =
						OutDataBuffer[bufferIndex].character<' '?' ':OutDataBuffer[bufferIndex].character;
				}
				else
				{
					// now, figure out if the field is protected or not 
					if ((OutDataBuffer[bufferIndex].field_attribute & CICS_3270_ATTR_PROT) > 0)
					{
						// protected field 
						scrline[x] = protChar;
					}
					else
					{
						// unprotected field 
						scrline[x] = unprotChar;
					}
				}

				bufferIndex++;
			}

			// end of screen line - move to next output line 
			scrline[x] =  '\n';
			scrline[x + 1] =  '\0';
			fprintf(DEBUG_FILE,scrline);
		}

		// calculate the cursor position 
		column = cursorPosition%numberColumns;
		row = cursorPosition/numberColumns;

		// display the cursor position and field counts 
		sprintf(scrline,"Cursor Position = %.2d,%.2d\n",row,column);
		fprintf(DEBUG_FILE,scrline);
		sprintf(scrline,"Total fields - %d  unprotected fields - %d  fields with MDT set - %d\n",
				numberFields, numberUnprotFields, MDTcount);

		fprintf(DEBUG_FILE,scrline);
   } // if (DEBUG)
}

void C3270::CICS_3270_GetProtTable(char *buff)

{
	int		x,y;			// work variables 
	int		bufferIndex;	// offset within the screen buffer 
	char	fldType = '0';

	// Start by finding out the type of the first position on the screen
	if (numberFields == 0) 
	{
		fldType = '0';	// Screen is unprotected
	}
	else
	{
		// Find the last attribute byte
		bufferIndex = numberRows*numberColumns - 1;
		while ((OutDataBuffer[bufferIndex].cell_type == CICS_3270_CELL_CHARACTER) &&
				(bufferIndex >= 0))
		{
			// back up to the previous field
			bufferIndex--;
		}

		if ((OutDataBuffer[bufferIndex].field_attribute & CICS_3270_ATTR_PROT) > 0)
		{
			fldType = '1';	// Protected
		}
		else
		{
			fldType = '0';	// Unprotected
		}
	}

	// start with the first position 
	bufferIndex = 0;

	for (y=0; y<numberRows; y++)
	{
		for (x=0; x<numberColumns; x++)
		{
			if (OutDataBuffer[bufferIndex].cell_type ==
					CICS_3270_CELL_CHARACTER)
			{
				// Indicate the type of field
				buff[bufferIndex] = fldType;
               
			}
			else
			{
				if ((OutDataBuffer[bufferIndex].field_attribute & CICS_3270_ATTR_PROT) > 0)
				{
					// Indicate a protected attribute byte
					buff[bufferIndex] = '3';
					fldType = '1';	// Protected
				}
				else
				{
					// Indicate a protected attribute byte
					buff[bufferIndex] = '2';	// Unprotected attribute
					fldType = '0';	// Unprotected
				}
			}

			bufferIndex++;
		}
	}

	// Terminate the string properly
	buff[bufferIndex] = '\0';

	return;
}

////////////////////////////////////////////////////////////////
//
// This routine will return the entire screen as a single 
// string.
//  
////////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetScreenImage(char* buff, char fieldCharProt, char fieldCharUnprot)

{
	int		x,y;			// work variables 
	int		bufferIndex;	// offset within the screen buffer 

	// start with the first position 
	bufferIndex = 0;

	for (y=0; y<numberRows; y++)
	{
		for (x=0; x<numberColumns; x++)
		{
			if (OutDataBuffer[bufferIndex].cell_type ==
					CICS_3270_CELL_CHARACTER)
			{
				// Display a blank if the data value is null
				buff[bufferIndex] = 
				 OutDataBuffer[bufferIndex].character<' '?' ':OutDataBuffer[bufferIndex].character;
               
			}
			else
			{
				if ((OutDataBuffer[bufferIndex].field_attribute & CICS_3270_ATTR_PROT) > 0)
				{
					// Display an "@" character for attribute byte locations
					buff[bufferIndex] = fieldCharProt;
				}
				else
				{
					// Display an "~" character for attribute byte locations
					buff[bufferIndex] = fieldCharUnprot;
				}
			}

			bufferIndex++;
		}

		// Insert a CR LF sequence
		//buff[bufferIndex++] = '\r';
		//buff[bufferIndex++] = '\n';
	}

	// Terminate the string properly
	buff[bufferIndex] = '\0';

	return;
}

void C3270::CICS_3270_FreeField(
/* IN */ CICS_3270_Field_t      *field)

{
        if (field->data != NULL)
        {
                free(field->data);
        }
        free(field);
        return;
}

void C3270::CICS_3270_GetField(
/* IN */        int row,
/* IN */        int column,
/* OUT */       CICS_3270_Field_t       **fieldRet)
{
   CICS_3270_Field_t    *field;
   int                  bufferIndex;
   int                  startIndex;
   int                  length;
   char                 tempChars[43*132];

   //
   // Allocate storage for returning the field
   //
   field = (CICS_3270_Field_t *)malloc(sizeof(CICS_3270_Field_t));

   //
   // Calculate the start buffer position
   //
   bufferIndex = row*numberColumns+column;
   startIndex = bufferIndex;

   //
   // We are now going to find the start of the field.  We do this by looking
   // for a cell with type=field.  We start at this position and work
   // backwards, wrapping from (0,0) to (79,23) if necessary.  The case where
   // there are no fields will be detected and a fake field is returned
   // (-1, -1)
   //
   
   while(1)
   {
   //
   // If we have found a field start, break the loop.
   //
      if (OutDataBuffer[bufferIndex].cell_type == CICS_3270_CELL_FIELD)
      {
         break;
      }
   //
   // back to the previous cell, wraping if necessary
   //
      PreviousCell(&bufferIndex);
   //
   // Check for the case where there are no fields ...
   //
      if (bufferIndex == startIndex)
      {
         SetNoField(field);
         *fieldRet = field;
         return;
      }
   }

   //
   // At this point, we have found the field start, start to fill in the
   // field structure for return.
   //
   field->row = bufferIndex/numberColumns;
   field->column = bufferIndex%numberColumns;
   field->attribute = OutDataBuffer[bufferIndex].field_attribute;
   field->extattr = OutDataBuffer[bufferIndex].ext_hilite;
   field->color = OutDataBuffer[bufferIndex].foregnd;

   //
   // We are now going to build the field data.  This is done by copying
   // characters until we find the next field terminator.
   //
   length = 0;
   while(1)
   {
      NextCell(&bufferIndex);
      if (OutDataBuffer[bufferIndex].cell_type == CICS_3270_CELL_FIELD)
      {
         break;
      }
      tempChars[length] = OutDataBuffer[bufferIndex].character;
      length++;
   }

   field->length = length;
   field->data = (char *) malloc(length+1);
   memcpy(field->data, tempChars, length);
   field->data[length] = 0;

   *fieldRet = field;
   return;
}

///////////////////////////////////////
//
// Routine to set the cursor position
//
///////////////////////////////////////

void C3270::CICS_3270_SetCursor(int row, int col)

{
   cursorPosition = row*numberColumns + col;
}

/////////////////////////////////////////////////
//
// Routine to print a field definition to stdout
//
/////////////////////////////////////////////////

void C3270::CICS_3270_PrintField(int rw, int cl)
{
   CICS_3270_Field_t   *fld;
   char            errline[MAX_LINE + 50];

   CICS_3270_GetField(rw, cl, &fld);

   // Display the results in trace file 
   sprintf(errline,"row %.2d col %.3d attr 0x%.2X lngth %.4d data len %.4d data %.50s\n",
               fld->row,fld->column,fld->attribute,fld->length,strlen(fld->data),fld->data);
   CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

   CICS_3270_FreeField(fld);
}

/////////////////////////////////////////////
//
// func(): GetFields(FIELD_STRUCT fields[])
//
///////////////////////////////////////

void C3270::GetFields(FIELD_STRUCT fields[])

{
	int i;
	int row;
	int col;
	int idx = 0;		// buffer index 
	int fldnum = 0;		// field number 

	// Next, clear the storage 
	for (i = 0; i <= numberFields; i++)
	{
		fields[i].bufofs = 0;		// buffer offset    
		fields[i].fldrow = 0;		// field row        
		fields[i].fldcol = 0;		// field column     
		fields[i].fldlen = 0;		// field length     
		fields[i].fldtype = 0;		// field type       
	}

	// Next, step through the buffer, looking for fields 
	for (row = 0; row < numberRows; row++)
	{
		for (col = 0; col < numberColumns; col++)
		{
			if (OutDataBuffer[idx].cell_type == CICS_3270_CELL_FIELD)
			{
				// Get the field location in the array entry 
				fields[fldnum].bufofs = idx;
				fields[fldnum].fldrow = row;
				fields[fldnum].fldcol = col;
				fields[fldnum].fldtype = OutDataBuffer[idx].field_attribute;

				// go on to the next field 
				fldnum++;
			}

			// Go on to the next screen position 
			idx++;
		}
	}

	// Lastly, calculate the lengths of the fields 
	for (i = 0; i < fldnum; i++)
	{
		if ( i == fldnum - 1)
		{
			// We must special case the last attribute on the screen 
			// calculate the length to the end of the screen buffer 
			fields[i].fldlen = (numberRows*numberColumns - 1) - fields[i].bufofs;

			// add the length at the beginning of the screen 
			fields[i].fldlen += fields[0].bufofs;
		}
		else
		{
			// length of this field is space between attribute bytes 
			fields[i].fldlen = fields[i + 1].bufofs - fields[i].bufofs - 1;
		}
	}
}

////////////////////////////////////////////////////////////////
//
// Routine to check if a character in a variable name is valid
// 
// To be valid, it must contain either a letter or a number
//
////////////////////////////////////////////////////////////////

int isLetter(char varLetter)

{
	if (((varLetter >= 'a') && (varLetter <= 'z')) ||
		((varLetter >= 'A') && (varLetter <= 'Z')) ||
		((varLetter >= '0') && (varLetter <= '9')))
	{
		// indicate valid 
		return 1;
	}
	else
	{
		// not a letter or number - indicate invalid 
		return 0;
	}
}

////////////////////////////////////////////////////////////////
//
// Routine to make sure the variable name is valid
// 
// Remove embedded spaces and make sure that the name only contains
// letters and numbers.
//
////////////////////////////////////////////////////////////////

void checkVarName(char *varname)

{
	size_t	i;

	i = 0;
	while (i < strlen(varname))
	{
		// check the character 
		if (!isLetter(varname[i]))
		{
			// the character is not a letter or number 
			// check for a single blank between valid characters 
			if ((varname[i] = ' ') && (isLetter(varname[i + 1])))
			{
				// change a single blank to an underscore 
				varname[i] = '_';
			}
			else
			{
				// truncate the variable name at this point 
				varname[i] = '\0';
			}
		}

		// go on to the next character 
		i++;
	}
}

void C3270::CICS_3270_WriteFields()

{
	FIELD_STRUCT *fields;
	int i;
	int j;
	int idx;
	short hash;
	char fldata[31];
	char temptype[31];
	int  templen;
	char	traceline[512];

	CICS_3270_WriteTraceLine("\n3270 Screen Definitions\n", CICS_3270_TRACE_MIN);

	if (numberFields == 0)
	{
		sprintf(traceline, "Screen is unformatted and contains no defined fields\n");
	}
	else
	{
		sprintf(traceline, "Num    Data                        row col len attributes\n\n");
		CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);

		// Allocate storage for the control blocks 
		fields = (_FIELD_STRUCT *) malloc((numberFields + 1) * sizeof(*fields));

		// build the field structure 
		GetFields(fields);

		for (i = 0; i < numberFields; i++)
		{
			// Clear the field data variable 
			memset(fldata, '\0', sizeof(fldata));

			// get the length of the field data, maximum of 30 characters 
			templen = fields[i].fldlen<(sizeof(fldata) - 1)?fields[i].fldlen:sizeof(fldata) - 1;

			// get the field data, if any 
			idx = fields[i].bufofs + 1;		// skip the attribute character 
			for (j = 0; j < templen; j++)
			{
				// Get the data from the buffer 
				fldata[j] = OutDataBuffer[idx].character;
				idx++;

				// change a null value to a blank 
				if (fldata[j] == 0)
				{
					fldata[j] = ' ';
				}
			}

			attr2str(fields[i].fldtype,temptype);
			
			sprintf(traceline, "%.3i %-30.30s %.3i %.3i %.3i %s\n", i + 1,
					fldata,
					fields[i].fldrow,
					fields[i].fldcol,
					fields[i].fldlen,
					temptype);

			// Write a trace entry
			CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
		}

		// release the allocated memory 
		free(fields);
	}

	// Get the attribute hash value
	hash = CICS_3270_ScreenSum();

	// Write a trace entry summarizing the screen
	sprintf(traceline,"Total fields %d  Unprotected fields %d  Attribute hash %d\n",
			numberFields, numberUnprotFields, hash);
	CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);

	return;
}

void C3270::CICS_3270_GetFieldAttrs(char* buff)

{
	FIELD_STRUCT *fields;
	int i;
	char temptype[31];
	char templine[80];
	int	 nextloc;

	// Make sure we have some fields defined
	if (numberFields == 0)
	{
		// No fields - just terminate the string
		buff[0] = '\0';
	}
	else
	{
		// Allocate storage for the control blocks 
		fields = (_FIELD_STRUCT *) malloc((numberFields + 1) * sizeof(*fields));

		// build the field structure 
		GetFields(fields);

		// Initialize the next field pointer
		nextloc = 0;

		// Loop through each defined field
		for (i = 0; i < numberFields; i++)
		{
			// Get the attribute byte as a string
			attr2str(fields[i].fldtype,temptype);
			
			// Get the field characteristics as a string
			sprintf(templine, "%.3i %.3i %.3i %.4i %-20.20s", i + 1,
					fields[i].fldrow,
					fields[i].fldcol,
					fields[i].fldlen,
					temptype);

			// Add the next entry to the total string
			strcat(buff + nextloc,templine);

			// Move on to the next field
			nextloc += strlen(templine);
		}

		// release the allocated memory 
		free(fields);
	}

	return;
}

///////////////////////////////
//
// func(): CICS_3270_GetCell()
//
///////////////////////////////

void C3270::CICS_3270_GetCell(
/* IN */        int row,
/* IN */        int column,
/* OUT */       CICS_3270_Cell_t *cell)

{
   int buffer;

   buffer = row * numberColumns + column;
   *cell = OutDataBuffer[buffer];
   return;
}

////////////////////////////////////
//
// Routine to simulate the tab key
//
////////////////////////////////////

void C3270::TabNextField(int *buffer, int direction)

{
   char         attribute;

   // Are there any unprotected fields? 
   if (numberUnprotFields)
   {
      while (1)
      {
         // Is this an attribute byte? 
         if (OutDataBuffer[*buffer].cell_type ==
             CICS_3270_CELL_FIELD)
         {
            // Is this an unprotected field? 
            attribute = OutDataBuffer[*buffer].field_attribute;
            if ((attribute & CICS_3270_ATTR_PROT) == 0)
            {
               NextCell(buffer);

               if (OutDataBuffer[*buffer].cell_type ==
                   CICS_3270_CELL_CHARACTER)
               {
                  return;
               }
            }
         }

         if (direction)
         {
            // move on to the next character 
            NextCell(buffer);
         }
         else
         {
            // move back to the next character 
            PreviousCell(buffer);
         } // endif 
      } // endwhile 
   }
   else
   {
      // No unprotected fields - set cursor to (0,0) 
      (*buffer) = 0;
   } // endif 
}

////////////////////////////////////////////////////
//
// Routine to move to the next unprotected location
//
////////////////////////////////////////////////////

void C3270::NextLocation(int *buffer)

{
   //
   // Move on to the next location
   //
   NextCell(buffer);

   //
   // If it is an attibute, skip past it to the next
   // unprotected field
   //

   if ((OutDataBuffer[*buffer].field_attribute &
       CICS_3270_ATTR_ASKIP) > 0)
   {
      TabNextField(buffer,1);
   }
}

////////////////////////////////////////////////////
//
// Routine to handle an auto skip attribute
//
////////////////////////////////////////////////////

void C3270::handleASKIP(int *buffer, int *setMDT)

{
   if (OutDataBuffer[*buffer].cell_type == CICS_3270_CELL_FIELD)
   {
      if ((OutDataBuffer[*buffer].field_attribute &
           CICS_3270_ATTR_ASKIP) == CICS_3270_ATTR_ASKIP)
      {
         if (numberUnprotFields > 0)
         {
            TabNextField(buffer,1);
            (*setMDT) = 0;
         } // endif 
      } // endif 
   } // endif 
}

////////////////////////////////////////////////////
//
// Routine to write error message for unrecognized
// escape sequences.
//
////////////////////////////////////////////////////

void C3270::badKeyMsg(const char *ch)

{
   char         errline[100];

   sprintf(errline,"  Bad escape sequence - &%c%c\n",*ch,*(ch + 1));
   CICS_3270_Write_Error_Log(errline);

   if (DEBUG)
   {
      CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
   } // endif 
}

////////////////////////////////////////////////////
//
// Routine to handle Delete Key
//
////////////////////////////////////////////////////

void C3270::deleteKey(int buffer, int *setMDT)

{
   int          curpos;
   int          nextpos;
   int          attrloc;

   if (OutDataBuffer[buffer].cell_type == CICS_3270_CELL_CHARACTER)
   {
      if (numberFields == 0)
      {
         nextpos = numberRows*numberColumns;
         for (curpos=buffer; curpos<nextpos; curpos++)
         {
            // set buffer location to null 
            OutDataBuffer[curpos].character = 0;
         } // endfor 
      }
      else
      {
         attrloc = buffer;
         findAttrChar(&attrloc);

         if ((OutDataBuffer[attrloc].field_attribute &
              CICS_3270_ATTR_PROT) == 0)
         {
            // get the current buffer position 
            curpos = buffer;
            nextpos = buffer;
            NextCell(&nextpos);

            // Move characters until attribute found 
            while (1)
            {
               // check if found attribute byte 
               if (OutDataBuffer[nextpos].cell_type ==
                   CICS_3270_CELL_FIELD)
               {
                  // set current character to null 
                  OutDataBuffer[curpos].character = 0;
                  break;
               }
               else
               {
                  // shift byte one to the left 
                  OutDataBuffer[curpos].character =
                     OutDataBuffer[nextpos].character;

                 // Move on to the next position 
                 curpos = nextpos;
                 NextCell(&nextpos);
               } // endif 
            } // endwhile 
         } // endif 
      } // endif 
   } // endif 
}

////////////////////////////////////////////////////
//
// Routine to handle Erase Input
//
////////////////////////////////////////////////////

void C3270::eraseInput()

{
   int buffer=0;
   int endbuf;
   int prot=1;

   if (numberFields = 0)
   {
      CICS_3270_ClearBuffer();
   }
   else
   {
      endbuf = numberRows*numberColumns;

      // find the attribute byte for (0,0)
      buffer = 0;
      if (OutDataBuffer[buffer].cell_type != CICS_3270_CELL_FIELD)
      {
         findAttrChar(&buffer);
		 if ((OutDataBuffer[buffer].field_attribute & CICS_3270_ATTR_PROT) == 0)
		 {
			 prot = 0;
		 }
      } // endif 

      buffer = 0;
      while(buffer < endbuf)
      {
         if (OutDataBuffer[buffer].cell_type == CICS_3270_CELL_FIELD)
         {
            if ((OutDataBuffer[buffer].field_attribute &
                CICS_3270_ATTR_PROT) == 0)
            {
               prot = 0; // field is unprotected 
            }
            else
            {
               prot = 1; // field is protected 
            } // endif 
         }
         else
         {
            // if unprotected field 
            if (prot = 0)
            {
               // replace the character data with a null 
               OutDataBuffer[buffer].character = 0;
            } // endif 
         } // endif 

         buffer++;
      } // endwhile 
   } // endif 
}

//////////////////////////////////////////////////////////
//
// Routine to write a character into the buffer.  The
// type of cell is checked to make sure that input is
// allowed.  The MDT for the field is set if this is the
// first write into this field.
//
//////////////////////////////////////////////////////////

short C3270::writeKeyStroke(char data, int *buffer, int *setMDT)

{
   int          attrloc;	// attribute location 
   short		rc = 0;		// return code
   char         attr;		// field attribute 
   char         errline[MAX_LINE + 50];

   if (numberFields > 0)
   {
      attrloc = *buffer;

      // Find the attribute byte for this field 
      findAttrChar(&attrloc);

      // get the attribute 
      attr = OutDataBuffer[attrloc].field_attribute;

      if (OutDataBuffer[(*buffer)].cell_type == CICS_3270_CELL_CHARACTER)
      {
         // check for unprotected field 

         if  ((attr & CICS_3270_ATTR_PROT) == 0)
         {
            // check for numeric field - either the field 
            // must be non-numeric or the data must be a  
            // number                                     
            if (((attr & CICS_3270_ATTR_NUMERIC) == 0) ||
                ((data >= '0') && (data <= '9')))
            {
               // OK to write the character 
               OutDataBuffer[(*buffer)].character = data;

               //
               // Set the modified data tag (MDT)
               //
               if (*setMDT == 0)
               {
                  // set the MDT in the field attribute  
                  OutDataBuffer[attrloc].field_attribute |=
                     CICS_3270_ATTR_MDT;

                  *setMDT = 1;  // MDT is set for this field 
               }

               //
               // Move on to the next location
               //
               NextCell(buffer);

               //
               // If it is an ASKIP attibute, skip past it
               // to the next unprotected field
               //

               handleASKIP(buffer, setMDT);
            }
            else
            {
				// set the return code to 2
				rc = 2;

				// log the error
               sprintf(errline,"Invalid write into numeric field - row %2d column %3d char %c\n",
                       (*buffer)/numberColumns,(*buffer)%numberColumns,data);
               CICS_3270_Write_Error_Log(errline);

               if (DEBUG)
               {
                  CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
               } // endif 
            } // endif 
         }
         else
         {
			 // set the return code to 1
			 rc = 1;

			 // log the error
            sprintf(errline,"Write into protected field - row %2d column %3d char %c\n",
                    (*buffer)/numberColumns,(*buffer)%numberColumns,data);
            CICS_3270_Write_Error_Log(errline);

            if (DEBUG)
            {
               CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
            } // endif 
         } // endif 
      }
      else
      {
		  // set the return code to 3
		  rc = 3;

		  // log the error
         sprintf(errline,"Attempt to type over attribute - row %2d column %3d char %c\n",
                 (*buffer)/numberColumns,(*buffer)%numberColumns,data);
         CICS_3270_Write_Error_Log(errline);

         if (DEBUG)
         {
            CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
         } // endif 
      }
   }
   else
   {
      // unformatted screen - just write the character 
      OutDataBuffer[(*buffer)].character = data;

      //
      // Move on to the next location
      //
      NextCell(buffer);
   } // endif 

   return rc;
}

//////////////////////////////////////////////////////////
//
// func(): CICS_3270_Keyin()
//
// This function will take a string of keystroke
// characters, and update the screen buffer as if
// the characters had been typed on a keyboard.
// Attention keys are not supported.  The INPUT
// command must be used for attention requests.
// This command is used as an alternative to MODFIELD.
// Certain special keys are defined as three character
// escape sequences, namely:
//
//  &Bn - Back Tab n times
//  &Dn - Cursor Left n times
//  &DL - Delete
//  &DU - Dup
//  &EF - Erase to end of field
//  &EI - Erase Input
//  &FM - Field Mark
//  &HO - Home
//  &Ln - Cursor Left n times
//  &Nn - Newline n times
//  &Rn - Cursor Left n times
//  &Tn - Tab n times
//  &Un - Cursor Left n times
//
//////////////////////////////////////////////////////////

short C3270::CICS_3270_Keyin(const char *cmdData)

{
   int          buffer;
   int          count;
   int          i;
   int          row;
   int          column;
   int          setMDT=0;
   short		rc = 0;		// return code
   short		rc1;
   char         ch;
   char			ch1;
   char			ch2;

   // initialize variables 
   buffer = cursorPosition;
   count = strlen(cmdData);

   while (count > 0)
   {
      // check for a escape character sequence 
      if ((cmdData[0] == '&') && (count >= 3) &&
          (cmdData[1] != '&'))
      {
         // convert a number in the last character to 
         // an integer                                

         if ((cmdData[2] >= '1') && (cmdData[2] <= '9'))
         {
            i = (int) cmdData[2] - (int) '0';
         }
         else
         {
            i =  0;     // not a number 
         } // endif 

         // convert the characters to upper case 
         ch1 = chrupr(cmdData[1]);
         ch2 = chrupr(cmdData[2]);

         // convert character to upper case 
         switch (ch1)
         {
            case 'B':
               if (i)
               {
                  // backtab i times 
                  while (i)
                  {
                     TabNextField(&buffer,0);
                     i--;
                  } // endwhile 
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  // set bad return code
				  rc = 4;
               } // endif 
               break;

            case 'D':
               if (i)
               {
                  // cursor down i times 
                  column = buffer%numberColumns;
                  row = buffer/numberColumns + i;
                  if (row >= numberRows)
                  {
                     row = row%numberRows;
                  }

                  buffer = row*numberColumns + column;
               }
               else
               {
                  if (ch2 == 'L')
                  {
                     // Delete key 
                     deleteKey(buffer,&setMDT);
                  }
                  else
                  {
                     if (ch2 == 'U')
                     {
                        // Dup key 
                        ch = (char) '\1c';
                        writeKeyStroke(ch,&buffer,&setMDT);
                     }
                     else
                     {
                        badKeyMsg(cmdData + 1);
		 				rc = 4;
                     } // endif 
                  } // endif 
               } // endif 
               break;

            case 'E':
               if (ch2 == 'F')
               {
                  // Erase to end of field 
                  i = buffer;  // start from current location 

                  while (1)
                  {
                     if ((OutDataBuffer[i].cell_type ==
                          CICS_3270_CELL_FIELD) ||
                         (i == numberRows*numberColumns))
                     {
                        break;
                     }
                     else
                     {
                        OutDataBuffer[i].character = 0;
                        i++;
                     } // endif 
                  } // endwhile 
               }
               else
               {
                  if (ch2 == 'I')
                  {
                     // Erase Input 
                     eraseInput();
                  }
                  else
                  {
                     badKeyMsg(cmdData + 1);
					  rc = 4;
                  } // endif 
               } // endif 
               break;

            case 'F':
               if (ch2 == 'M')
               {
                  // Field Mark 
                  ch = (char) '\1e';
                  writeKeyStroke(ch,&buffer,&setMDT);
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  rc = 4;
               }
               break;

            case 'H':
               if (ch2 == 'O')
               {
                  // Home 
                  // Move cursor position to 0,0 
                  buffer = 0;

				  // Check for an autoskip attribute character
				  // If an autoskip attribute is found, tab to the next
				  // unprotected field
				  if ((OutDataBuffer[buffer].cell_type == CICS_3270_CELL_FIELD) &&
					  ((OutDataBuffer[buffer].field_attribute & CICS_3270_ATTR_ASKIP) ==
					    CICS_3270_ATTR_ASKIP))
				  {
					  TabNextField(&buffer,1);
				  }
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  rc = 4;
               }

               break;

            case 'L':
               if (i)
               {
                  // cursor left i times 
                  buffer -= i;

                  // check if we wrapped 
                  if (buffer < 0)
                  {
                     buffer += numberColumns*numberRows;
                  } // endif 
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  rc = 4;
               } // endif 

               break;

            case 'N':
               if (i)
               {
                  // newline i times 
                  column = 0;
                  row = buffer/numberColumns + i;
                  if (row >= numberRows)
                  {
                     row = row%numberRows;
                  }

                  buffer = row*numberColumns + column;
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  rc = 4;
               } // endif 

               break;

            case 'R':
               if (i)
               {
                  // cursor right i times 
                  buffer += i;
                  if (buffer >= (numberColumns*numberRows))
                  {
                     buffer -= numberColumns*numberRows;
                  } // endif 
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  rc = 4;
               } // endif 

               break;

            case 'T':
               if (i)
               {
                  // tab i times 
                  while (i)
                  {
                     TabNextField(&buffer,1);
                     i--;
                  } // endwhile 
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  rc = 4;
               } // endif 

               break;

            case 'U':
               if (i)
               {
                  // cursor up i times 
                  column = buffer%numberColumns;
                  row = buffer/numberColumns - i;
                  if (row < 0)
                  {
                     row += numberRows;
                  }

                  buffer = row*numberColumns + column;
               }
               else
               {
                  badKeyMsg(cmdData + 1);
				  rc = 4;
               } // endif 

               break;

            default:
                badKeyMsg(cmdData + 1);
				  rc = 4;
                break;

         } // endswitch cmdData[1] 

         // May be new fields, so must set MDT 
         setMDT = 0;

         // move past the two characters and escape character 
         cmdData += 3;
         count -= 3;
      }
      else
      {
         if ((cmdData[0] == '&') && (cmdData[1] == '&'))
         {
            cmdData++; // skip one of the ampersands 
            count--;
         }

         // Normal character - write to screen 
         rc1 = writeKeyStroke(cmdData[0],&buffer,&setMDT);
		 // check for error
		 if (rc == 0)
		 {
			 rc = rc1;
		 }

         cmdData++;
         count--;
      } // endif cmdData[0] == '&' 

   } // endwhile count > 0 

   // remember the new cursor position 
   cursorPosition = buffer;

   return rc;
}

/////////////////////////////////////////////////////////////
//
// func(): CICS_3270_ModifyField()
//
/////////////////////////////////////////////////////////////

short C3270::CICS_3270_ModifyField(
/* IN */ int row,
/* IN */ int column,
/* IN */ const char *data)

{
        int			buffer;
		char		errline[MAX_LINE + 50];
		short		rc=0;	// return code

		// return if no fields on screen
		if (numberFields == 0)
		{
			// Build the error message
			sprintf(errline,"3 ModifyField - No fields defined on screen\n");

			// Write a trace message
			CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// Write Error Log record
			CICS_3270_Write_Error_Log(errline);

			// Indicate no fields on screen
			return 3;
		}

		// Validate the row and column
		if ((row < 0) || (row >= numberRows) || (column < 0) || (column >= numberColumns))
		{
			// Build the error message
			sprintf(errline,"1 ModifyField - Invalid row %d or column %d\n",row,column);

			// Write a trace message
			CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// Write Error Log record
			CICS_3270_Write_Error_Log(errline);

			// Indicate an invalid row or column 
			return 1;
		}

		// Calculate the buffer offset
		buffer = row * numberColumns + column;

		// Check for an attribute byte at this location
		if (OutDataBuffer[buffer].cell_type == CICS_3270_CELL_FIELD)
		{
			// Check if the field is protected
			if ((OutDataBuffer[buffer].field_attribute & CICS_3270_ATTR_PROT) == 0)
			{
		        // Set the modified data tag (MDT)
				OutDataBuffer[buffer].field_attribute |= CICS_3270_ATTR_MDT;

			    // Actual insert starts AFTER the attribute
				NextCell(&buffer);

				// We must now loop until we hit the end of the data or
				// the next attribute byte
				while ((data[0] != '\0') &&
						(OutDataBuffer[buffer].cell_type == CICS_3270_CELL_CHARACTER))
				{
					// Update the data at the screen location
					OutDataBuffer[buffer].character = data[0];

					// move to the next screen location, wrapping if necessary
					NextCell(&buffer);

					// get the next input character
					data++;
				}

				// Lastly, pad the field with null characters if necessary, until
				// we find the next field
				while (OutDataBuffer[buffer].cell_type == CICS_3270_CELL_CHARACTER)
				{
					// Update the data at the screen location
					OutDataBuffer[buffer].character = '\0';

					// move to the next screen location, wrapping if necessary
					NextCell(&buffer);
				}
			}
			else
			{
				// Error - Field is protected
				sprintf(errline,"2 ModifyField - Field at row %d column %d is protected\n",row,column);

				// Write a trace message
				CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

				// Write Error Log record
				CICS_3270_Write_Error_Log(errline);

				// Indicate field is protected
				return 2;
			}
		}
		else
		{
			// This location is not an attribute byte
			// Return an error message - Invalid row or column
			sprintf(errline,"1 ModifyField - No field at row %d column %d\n",row,column);

			// Write a trace message
			CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// Write Error Log record
			CICS_3270_Write_Error_Log(errline);

			// Indicate an invalid row or column 
			return 1;
		}

		// normal return
        return 0;
}

/////////////////////////////////////////////////////////////
//
// func(): CICS_3270_GetInbound()
//
/////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetInbound(
/* IN */        char aid,
/* OUT */       char **data,
/* OUT */       int *length)

{
   unsigned short  ascii;
   int             bufferPtr;
   int             outPtr;
   int             tempPtr;
   int             oldBufferPtr;
   char            Inbound[256*256];
   char            errline[MAX_LINE + 50];

   //
   // The inbound data stream looks as follows:
   //
   //    0           1,2        3      4,5     ...
   // +-----+----------------+-----+---------+------+------
   // | AID | Cursor address | SBA | Address | data | ...
   // +-----+----------------+-----+---------+------+------
   //
   //

   // Set the inbound aid byte in the data stream
   Inbound[0] = aid;

   // Set the cursor address in the inbound data stream
   CICS_3270_SBA_BufferToAscii(cursorPosition, &ascii);
   *(short *)(Inbound+1) = ascii;

   bufferPtr = 0;
   outPtr = 3;

   // check for formatted screen 
   if (numberFields == 0)
   {
	   // Unformatted screen - stop at the first null character
      while ((OutDataBuffer[bufferPtr].character != '\0') &&
		     (bufferPtr < cursorPosition) &&
             (bufferPtr < numberRows*numberColumns ))
      {
         // move the data character to the output buffer 
         Inbound[outPtr] = OutDataBuffer[bufferPtr].character;

         // increment the pointers 
         bufferPtr++;
         outPtr++;
      }

      if (DEBUG)
      {
         CICS_3270_WriteTraceLine("Buffer is unformatted\n", CICS_3270_TRACE_SOME);
      }
   }
   else
   { // formatted screen 
      while (1)
      {
         if (OutDataBuffer[bufferPtr].cell_type == CICS_3270_CELL_FIELD)
         {
      //
      // Start of field
      //

            if (OutDataBuffer[bufferPtr].field_attribute & CICS_3270_ATTR_MDT)
            {
      //
      // MDT bit set
      //
               if ((DEBUG) && (DEBUG_LEVEL == CICS_3270_TRACE_MAX))
               {
                  fprintf(DEBUG_FILE, "MDT> Found at (%d, %d)\n",
                          bufferPtr/numberColumns, bufferPtr%numberColumns);
               }


               Inbound[outPtr] = CICS_3270_ORDER_SBA;
               outPtr++;
               tempPtr = bufferPtr;
               NextCell(&tempPtr);
      //
      // Convert to ASCII for data stream
      //
               CICS_3270_SBA_BufferToAscii(tempPtr, &ascii);
               *(short *)(Inbound+outPtr) = ascii;
               outPtr+=2;

      //
      // Loop through all the characters
      //
               while (1)
               {
      //
      // Terminate when we find the next field marker or hit a null
      // character
      //
                  if ((OutDataBuffer[tempPtr].cell_type
                          == CICS_3270_CELL_FIELD) ||
                      (OutDataBuffer[tempPtr].character == 0))
                  {
                     break;
                  }
      //
      // Copy the character and advance the pointers
      //
                  Inbound[outPtr] = OutDataBuffer[tempPtr].character;
                  outPtr++;
                  NextCell(&tempPtr);
               } // End of character data processing 
            } // End of MDT field processing 
         } // End of field processing 

   //
   // Access the next cell, if we have wrapped, we have gone through the
   // whole buffer looking for fields.
   //
         oldBufferPtr = bufferPtr;
         NextCell(&bufferPtr);
         if (oldBufferPtr > bufferPtr)
         {
            break;
         }
      }
   } // endif formatted screen 

   //
   // Finished building the internal inbound buffer,
   // time to malloc and return the copy.
   //

   *data = (char *)malloc(outPtr);
   *length = outPtr;
   memcpy(*data, Inbound, outPtr);

   if (DEBUG)
   {
	   // Write a trace entry
      sprintf(errline, "\nInbound Data Stream Built - %d bytes\n", outPtr);
	  CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_SOME);

	  // write the raw data stream to the trace file
	  if (DEBUG_LEVEL == CICS_3270_TRACE_MAX)
	  {
		  dumpDatastream((*data), outPtr);
	  }
   } // endif 

   return;
}

//////////////////////////////////////////////////////////////////
//
// func(): CICS_3270_GetBuffer()
//
// This routine is used for read buffer requests.  These
// requests are generated by an EXEC CICS RECEIVE BUFFER command.
//
//////////////////////////////////////////////////////////////////

void C3270::CICS_3270_GetBuffer(
/* IN */        char aid,
/* OUT */       char **data,
/* OUT */       int *length)
{
   unsigned short  ascii;
   int             bufferPtr;
   int             outPtr;
   int             tempPtr;
   int             oldBufferPtr;
   int             i;
   char            tempAttr[20];  // attribute type/value pairs 
   char            Inbound[256*256];
   char            errline[MAX_LINE + 50];

   if (DEBUG)
   {
      CICS_3270_WriteTraceLine("\nBuilding response for Read Buffer\n", CICS_3270_TRACE_MIN);
   }

   //////////////////////////////////////////////////////////////
   //
   // The inbound data stream looks as follows:
   //
   //    0     1,2      3      4,5     6     7      8-n   ...
   // +-----+----------------+-----+---------+------+----------
   // | AID | Cursor | SBA | Address | sf | attr | data |
   // +-----+----------------+-----+---------+------+----------
   //
   //////////////////////////////////////////////////////////////

   Inbound[0] = aid;

   CICS_3270_SBA_BufferToAscii(cursorPosition, &ascii);
   *(short *)(Inbound+1) = ascii;

   bufferPtr = 0;
   outPtr = 3;

   // check for formatted screen 
   if (numberFields == 0)
   {
      while (bufferPtr < numberRows*numberColumns )
      {
         // move the data character to the output buffer 
         Inbound[outPtr] = OutDataBuffer[bufferPtr].character;

         // increment the pointers 
         bufferPtr++;
         outPtr++;
      }

      if ((DEBUG) && (DEBUG_LEVEL >= CICS_3270_TRACE_SOME))
      {
         CICS_3270_WriteTraceLine("Buffer is unformatted\n", CICS_3270_TRACE_SOME);
      }
   }
   else
   { // formatted screen 
      while (1)
      {
         if (OutDataBuffer[bufferPtr].cell_type == CICS_3270_CELL_FIELD)
         {
         //
         // Start of field - write to trace file if needed
         //

            if ((DEBUG) && (DEBUG_LEVEL >= CICS_3270_TRACE_SOME))
            {
               fprintf(DEBUG_FILE, "FIELD> Found at (%d, %d)\n",
                       bufferPtr/numberColumns, bufferPtr%numberColumns);
            }

         //
         // first build the SBA sequence
         //
            Inbound[outPtr] = CICS_3270_ORDER_SBA;
            outPtr++;
            tempPtr = bufferPtr;  // initialize data pointer 

         //
         // Add the address to complete the SBA order,
         // converting to ASCII for data stream
         //
            CICS_3270_SBA_BufferToAscii(tempPtr, &ascii);
            *(short *)(Inbound+outPtr) = ascii;
            outPtr+=2;

         //
         // check for any extended attributes
         //
            if (checkExtAttr(bufferPtr) == 0)
            {
         //
         // no extended attributes - use Start Field (SF)
         //
               Inbound[outPtr] = CICS_3270_ORDER_SF;
               Inbound[outPtr + 1] =
                  Bits2Ascii[OutDataBuffer[bufferPtr].field_attribute];
               outPtr += 2;
            }
            else
            {
         //
         // extended attributes - use Start Field Extended (SFE)
         //
               Inbound[outPtr] = CICS_3270_ORDER_SFE;

               // build the type/value pairs for this location 
               buildAttrPairs(bufferPtr, tempAttr);

               // insert the count of type/value pairs 
               Inbound[outPtr + 1] = strlen(tempAttr) / 2;

               // insert the type/value pairs  
               memcpy(&Inbound[outPtr + 2], tempAttr, strlen(tempAttr));

               // add the field attribute type/value pair 
               outPtr += 2 + strlen(tempAttr);
            } // endif 

            NextCell(&tempPtr);   // point to the first data byte 

         //
         // Loop through all the characters
         //
            while(1)
            {
         //
         // Terminate when we find the next field marker
         //
               if ((OutDataBuffer[tempPtr].cell_type
                       == CICS_3270_CELL_FIELD) ||
                   (OutDataBuffer[tempPtr].character == '\0'))
               {
                  break;
               }
         //
         // Check for character attributes
         //
               if (checkExtAttr(bufferPtr) == 0)
               {
                  // build the type/value pairs for this location 
                  buildAttrPairs(tempPtr, tempAttr);

                  for (i=0; (unsigned int) i < strlen(tempAttr);)
                  {
                     Inbound[outPtr++] = CICS_3270_ORDER_SA;
                     Inbound[outPtr++] = tempAttr[i++];
                     Inbound[outPtr++] = tempAttr[i++];
                  } // endfor 
               }

         //
         // Copy the character and advance the pointers
         //
               Inbound[outPtr] = OutDataBuffer[tempPtr].character;
               outPtr++;
               NextCell(&tempPtr);
            } // End of character data processing 
         } // End of field processing 

   //
   // Access the next cell, if we have wrapped, we have gone through the
   // whole buffer looking for fields.
   //
         oldBufferPtr = bufferPtr;
         NextCell(&bufferPtr);
         if (oldBufferPtr > bufferPtr)
         {
            break;
         }
      }  // while (1) 
   } // formatted screen 

   //
   // Finished building the internal inbound buffer,
   // time to malloc and return the copy.
   //

   *data = (char *)malloc(outPtr);
   *length = outPtr;
   memcpy(*data, Inbound, outPtr);

   if (DEBUG)
   {
	   // Write a trace entry
      sprintf(errline, "\nBuffer Data Stream Built - %d bytes\n", outPtr);
	  CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	  // write the raw data stream to the trace file
	  if (DEBUG_LEVEL == CICS_3270_TRACE_MAX)
	  {
		dumpDatastream((*data), outPtr);
	  }
   } // endif 

   return;
}

///////////////////////////////////////////////////////////
//
// This routine writes the complete attributes of each
// field into the trace file.  
//
///////////////////////////////////////////////////////////

void C3270::CICS_3270_DumpFields()
{
   int     x,y,i;
   int     bufferIndex;
   char    protectfld[50];
   char		traceline[512];

   CICS_3270_WriteTraceLine("\nDump of screen fields \n\n", CICS_3270_TRACE_MIN);

   bufferIndex = 0;
   for (y=0; y<numberRows; y++)
   {
      for (x=0; x<numberColumns; x++)
      {
         if (OutDataBuffer[bufferIndex].cell_type ==
                 CICS_3270_CELL_FIELD)
         {
            i = OutDataBuffer[bufferIndex].field_attribute;
            attr2str(i,protectfld);

            sprintf(traceline, "row %.2d column %.3d attribute 0x%.2X %s\n",
              y, x, i, protectfld);
			CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);

            i = OutDataBuffer[bufferIndex].foregnd;
            if (i > 0)
            {
               color2string(OutDataBuffer[bufferIndex].foregnd,protectfld);
               sprintf(traceline, "          Foreground Color 0x%.2X         %s\n",
                       i, protectfld);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 

            i = OutDataBuffer[bufferIndex].backgnd;
            if (i > 0)
            {
               color2string(OutDataBuffer[bufferIndex].backgnd,protectfld);
               sprintf(traceline, "          Background Color 0x%.2X         %s\n",
                       i, protectfld);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 

            i = OutDataBuffer[bufferIndex].ext_hilite;
            if (i > 0)
            {
               exthi2string(OutDataBuffer[bufferIndex].ext_hilite,protectfld);
               sprintf(traceline, "          Extended Hilite  0x%.2X         %s\n",
                       i, protectfld);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 

            i = OutDataBuffer[bufferIndex].field_validate;
            if (i > 0)
            {
               sprintf(traceline, "          Field Validate   0x%.2X\n",i);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 

            i = OutDataBuffer[bufferIndex].fld_outline;
            if (i > 0)
            {
               sprintf(traceline, "          Field Outline    0x%.2X\n",i);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 

            i = OutDataBuffer[bufferIndex].ps_char_set;
            if (i > 0)
            {
               sprintf(traceline, "          PS Character Set 0x%.2X\n",i);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 

            i = OutDataBuffer[bufferIndex].transparency;
            if (i > 0)
            {
               sprintf(traceline, "          Transparency     0x%.2X\n",i);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 

            i = OutDataBuffer[bufferIndex].SOSI;
            if (i > 0)
            {
               sprintf(traceline, "          SOSI             0x%.2X\n",i);
				CICS_3270_WriteTraceLine(traceline, CICS_3270_TRACE_MIN);
            } // endif 
         }

         bufferIndex++;
      }

   }

   return;
}

// This routine creates a hash sum for the screen, by adding the location
// of each attribute byte, and truncating to a 16-bit integer.  The purpose
// of this hash sum is for an application to check if it has received the
// screen it is expecting.

int C3270::CICS_3270_ScreenSum()

{
	long	sum=0;
	int		i;
	
	if (numberFields > 0)
	{
		for (i=0; i < numberRows*numberColumns; i++)
		{
			if (OutDataBuffer[i].cell_type == CICS_3270_CELL_FIELD)
			{
				sum += i + 1;
			}
		}
	}

	return (sum & 0x0000FFFF);
}

int C3270::CICS_3270_getNumberFields()

{
	return numberFields;
}

int C3270::CICS_3270_getNumberUnprotFields()

{
	return numberUnprotFields;
}

///////////////////////////////////////////////////
//
// Routine to get the number of rows and columns
// 
///////////////////////////////////////////////////

void C3270::CICS_3270_GetScreenParms(short* rows, short* cols)

{
	// set the number of rows and columns
	*rows = numberRows;
	*cols = numberColumns;
}

/////////////////
// End Of File //
/////////////////
